/* 
 * thinc_zsound.h: header for sound encoding library
 *
 * Authors: 
 * Ricardo A. Baratto - Network Computing Lab - Columbia University
 */

/* the idea with this lib is to provide a library that can abstract the details
 * of dealing with a particular sound encoding library. which lib to use
 * is set up at initialization time, and then it's all good from there.
 * hope is to support ogg vorbis (for normal sound) and speex (for voice data).
 * hopefully the model will work for others, tho using too many may get messy.
 */

/* set tabstop = 8 */

#ifndef _THINC_ZSOUND_H_
#define _THINC_ZSOUND_H_

#include <stdarg.h>

/* which backend to use */
#define T_ZSOUND_VORBIS	0x01
#define T_ZSOUND_SPEEX	0x02

/* some return values (we already use T_OK and T_ERROR) */
#define T_ZSOUND_DONE	0

/* same for encode/decode */
struct thinc_zsound {
	void *handle;	/* one of _vorbis or _speex */

	/* callback functions for writing data out */
	int (*write)(void *handle, const void *buffer, unsigned int len);
	int (*read)(void *handle, void *buffer, unsigned int len);
	void *cb_handle;

	/* and some function pointers to make this look all 
	 * professional */
	int (*init)(struct thinc_zsound*, va_list params);
	void (*end)(struct thinc_zsound*);
	int (*prepare)(struct thinc_zsound*);
	int (*encode)(struct thinc_zsound*, char *, unsigned int);
	int (*decode)(struct thinc_zsound*, char *, int);
	int (*finish)(struct thinc_zsound*);
};

/***************** encoder *****************/
/* set things up for an encoder run */
struct thinc_zsound* thinc_zsound_enc_init(int which, ...);
/* call this after init and before doing ANYTHING else */
void thinc_zsound_enc_setcb(struct thinc_zsound *, 
	int (*write)(void *, const void *, unsigned int), void *handle);
/* call this right before you start encoding data */
int thinc_zsound_enc_prepare(struct thinc_zsound *);
int thinc_zsound_enc_encode(struct thinc_zsound*, char *, 
	unsigned int);
/* done encoding, flush all the data */
int thinc_zsound_enc_finish(struct thinc_zsound*);
/* done, clean up time (for now we can't reuse across multiple encoding
 * calls so you'll need to call end/init before you can start encoding 
 * again) */
void thinc_zsound_enc_end(struct thinc_zsound*);

/***************** decoder *****************/
/* the decoder differs from the encoder in that you only have to init once
 * and then just call prepare,decode,finish multiple times and it'll work
 * just fine (tm) */
/* initialize */
struct thinc_zsound* thinc_zsound_dec_init(int which, ...);
/* call this after init and before doing ANYTHING else */
void thinc_zsound_dec_setcb(struct thinc_zsound *zs, 
	int (*read)(void *, void *, unsigned int), 
	void *cb_handle);
/* call this right before you start decoding data */
int thinc_zsound_dec_prepare(struct thinc_zsound *);
/* CAREFUL, weird semantics. decode some data and put it in the passed buffer. 
 * won't decode more than len bytes. how much data was decoded is returned, 
 * or T_ZSOUND_DONE if we're at the end of the stream (and you should call 
 * finish()). the read() callback will be used to read in data needed
 * to decode */
int thinc_zsound_dec_decode(struct thinc_zsound*, char* decoded, int len);
/* done decoding, flush all the data */
int thinc_zsound_dec_finish(struct thinc_zsound*);
/* really done, no more work to do. note you can do multiple decoding runs
 * without having to call init/finish repeatedly */
void thinc_zsound_dec_end(struct thinc_zsound*);

#endif
