/* Copyright 2005, 2007 Ricardo Baratto */
/* Copyright 2005, 2007 Columbia University */
/* This file is part of THINC.

    THINC is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 2 of the License, or
    (at your option) any later version.

    THINC is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with THINC.  If not, see <http://www.gnu.org/licenses/>.
*/
/* 
 * thinc_zsound_enc.c: routines for encoding sound
 *
 * Authors: 
 * Ricardo A. Baratto - Network Computing Lab - Columbia University
 */

/* set tabstop = 8 */

#include "thinc_zsound.h"
#include "thincCommon.h"
#include <vorbis/vorbisenc.h>
#include <stdlib.h>
#include <time.h>
#include <math.h>

#define __TZ_IN_LIB
#include "thinc_zsound_priv.h"

/* sound encoding library specific structs */
struct tz_vorbis {
	vorbis_info vi;	/* struct that stores all the static vorbis bitstream
			   settings */
	vorbis_dsp_state vd; 	/* central working state for the packet->PCM 
				   decoder */
	vorbis_block vb;	/* local working space for packet->PCM decode*/
	vorbis_comment vc;
	ogg_stream_state os;	/* take physical pages, weld into a logical
				   stream of packets */
	/* used to convert from bytes to channels per sample (unit used by
	 * libvorbis) */
	unsigned int shift_channels;
};

struct tz_speex {
		/* nothing for now as we have no speex support */
};

tz_funcs_enc(vorbis);
tz_funcs_enc(speex);

#define tz_fill_funcs(zs, name)			\
	zs->init = tz_init_##name;		\
	zs->end = tz_end_##name;		\
	zs->prepare = tz_prepare_##name;	\
	zs->encode = tz_encode_##name;	\
	zs->finish = tz_finish_##name;

/* other local functions */
static int tz_vorbis_doit(struct thinc_zsound *zs);

/* real code starts here */
struct thinc_zsound* thinc_zsound_enc_init(int which, ...)
{
	va_list params;
	struct thinc_zsound *zs;
	int err;

	zs = malloc(sizeof(*zs));
	if(!zs)
		return NULL;

	switch(which) {
		case T_ZSOUND_VORBIS:
			tz_fill_funcs(zs, vorbis);
			break;
		case T_ZSOUND_SPEEX:
			tz_fill_funcs(zs, speex);
			break;
		default:
			free(zs);
			return NULL;
	}

	va_start(params, which);
	err = zs->init(zs, params);
	va_end(params);
	if(err == T_ERROR) {
		free(zs);
		return NULL;
	}
	zs->read = NULL;
	zs->write = NULL;
	zs->cb_handle = NULL;

	return zs;
}

/* call this after init and before doing ANYTHING else */
void thinc_zsound_enc_setcb(struct thinc_zsound *zs, 
	int (*write)(void *, const void *, unsigned int), void *cb_handle)
{
	zs->write = write;
	zs->cb_handle = cb_handle;
}

void thinc_zsound_enc_end(struct thinc_zsound* zs)
{
	zs->end(zs);
	free(zs);
}

int thinc_zsound_enc_prepare(struct thinc_zsound *zs)
{
	return zs->prepare(zs);
}
int thinc_zsound_enc_encode(struct thinc_zsound* zs, char *buffer, 
	unsigned int len)
{
	return zs->encode(zs, buffer, len);
}

/* done encodeing, flush all the data */
int thinc_zsound_enc_finish(struct thinc_zsound* zs)
{
	return zs->finish(zs);
}

/**** vorbis funcs ****/
/* most of this code taken from encoder_example.c from libvorbisenc dist.
 * thanks guys! */

/* Init parameters (va_list) are info about the sound, and then the encoding 
 * quality:
 * long channels
 * long rate
 * double quality (libvorbis uses float but gcc promotes float to double
 * when passing through va_list and we don't like compiler warnings)
 *
 * about the quality, we encode using a VBR quality mode.  The usable 
 * range is -.1 (lowest quality, smallest file) to 1. (highest quality, 
 * largest file).
 * Example quality mode .4: 44kHz stereo coupled, roughly 128kbps VBR 
 */
static int tz_init_vorbis(struct thinc_zsound *zs, va_list params)
{
	struct tz_vorbis *vorbis;
	long channels;
	long rate;
	double quality;
	int err;

	vorbis = malloc(sizeof(*vorbis));
	if(!vorbis)
		return T_ERROR;

	channels = va_arg(params, long);
	rate = va_arg(params, long);
	quality = va_arg(params, double);

	TZ_DEBUG("encoding: sound params: %ld, %ld, %f\n", channels, rate, 
		quality);

	/* XXX: blindly assuming that it's always 16bit per channel. is that
	 * valid (note that caller is not telling us that info anyways) */
	switch(channels) {
		case 1:
			/* /2 */
			vorbis->shift_channels = 1;
			break;
		case 2:
			/* /4 */
			vorbis->shift_channels = 2;
			break;
		default:
			/* Add more as needed, i just don't know
			 * which other configs may be used */
			free(vorbis);
			return T_ERROR;
	}

	vorbis_info_init(&vorbis->vi);
	err = vorbis_encode_init_vbr(&vorbis->vi, channels, rate, quality);
	if(err) {
		vorbis_info_clear(&vorbis->vi);
		free(vorbis);
		return T_ERROR;
	}

	/* add a dummy comment */
	vorbis_comment_init(&vorbis->vc);
	vorbis_comment_add_tag(&vorbis->vc, "ENCODER", "THINC");

	/* set up analysis state and auxiliary encoding storage */
	vorbis_analysis_init(&vorbis->vd, &vorbis->vi);
	vorbis_block_init(&vorbis->vd, &vorbis->vb);

	/* set up our packet->stream encoder */
	/* pick a random serial number; that way we can more likely build
	   chained streams just by concatenation */
	srand(time(NULL));
	ogg_stream_init(&vorbis->os, rand());

	zs->handle = (void*) vorbis;
	
	return T_OK;
}

/* you should've set up the callbacks before calling this function */
static int tz_prepare_vorbis(struct thinc_zsound *zs)
{
	struct tz_vorbis *vorbis = zs->handle;
	/* Vorbis streams begin with three headers; the initial header (with
	 * most of the codec setup parameters) which is mandated by the Ogg
	 * bitstream spec.  The second header holds any comment fields.  The
	 * third header holds the bitstream codebook.  We merely need to make
	 * the headers, then pass them to libvorbis one at a time; libvorbis
	 * handles the additional Ogg bitstream constraints.
	 * But for this case, we don't have comment so we just skip over that */
	ogg_packet header;
	ogg_packet header_code;
	ogg_page og;
	ogg_packet header_comm;

	vorbis_analysis_headerout(&vorbis->vd, &vorbis->vc, &header, 
		&header_comm, &header_code);
	/* automatically placed in its own page */ 
	ogg_stream_packetin(&vorbis->os, &header);	
	ogg_stream_packetin(&vorbis->os, &header_comm);
	ogg_stream_packetin(&vorbis->os, &header_code);

	/* This ensures the actual audio data will start on a new 
	 * page, as per spec */
	do {
		if(ogg_stream_flush(&vorbis->os, &og) == 0)
			break;

		zs->write(zs->cb_handle, og.header, og.header_len);
		zs->write(zs->cb_handle, og.body, og.body_len);
	} while(1);

	return T_OK;
}

/* len is in bytes, which we convert to "samples per channel" as that's what
 * libvorbis expects. so make sure that len is multiple of channels * 2 */
static int tz_encode_vorbis(struct thinc_zsound *zs, char *buffer,
	unsigned int len)
{
	struct tz_vorbis *vorbis = zs->handle;
	unsigned int i, samples;
	float **pcm;

	samples = len >> vorbis->shift_channels; 
	pcm = vorbis_analysis_buffer(&vorbis->vd, samples);

	/* uninterleave samples */
	/* XXX: this seems kind of crazy, do we always have to do it?? (and
	 * for all data formats??) */
	for (i = 0; i < samples; ++i) {
		pcm[0][i] = 
			(
			 (buffer[i * 4 + 1] << 8) |
			 (0x00ff & (int) buffer[i * 4])
			) / 32768.f;
		pcm[1][i] =
			(
			 (buffer[i * 4 + 3] << 8) |
			 (0x00ff & (int)buffer[i * 4 + 2])
			) / 32768.f;
	}
	vorbis_analysis_wrote(&vorbis->vd, i);

	return tz_vorbis_doit(zs);
}

static int tz_finish_vorbis(struct thinc_zsound *zs)
{
	struct tz_vorbis *vorbis = zs->handle;

	vorbis_analysis_wrote(&vorbis->vd, 0);

	return tz_vorbis_doit(zs);
}

static void tz_end_vorbis(struct thinc_zsound *zs)
{
	struct tz_vorbis *vorbis = zs->handle;

	ogg_stream_clear(&vorbis->os);
	vorbis_block_clear(&vorbis->vb);
	vorbis_dsp_clear(&vorbis->vd);
	vorbis_comment_clear(&vorbis->vc);
	vorbis_info_clear(&vorbis->vi);
	free(vorbis);
	zs->handle = NULL;
}

/* helper function, used by encode and finish, as there doesn't seem to
 * be any obvious difference between the two cases [except for telling libvorbis
 * the amount of data available] (probably i need to understand libvorbis code 
 * better, but who cares) */
static int tz_vorbis_doit(struct thinc_zsound *zs)
{
	struct tz_vorbis *vorbis = zs->handle;
	vorbis_dsp_state *vd = &vorbis->vd;
	vorbis_block *vb = &vorbis->vb;
	ogg_stream_state *os = &vorbis->os;
	ogg_packet op;
	ogg_page og;

	/* vorbis does some data preanalysis, then divvies up blocks for
	   more involved (potentially parallel) processing.  Get a single
	   block for encoding now */
	while (vorbis_analysis_blockout(vd, vb) == 1) {
		/* analysis, assume we want to use bitrate management */
		vorbis_analysis(vb, NULL);
		vorbis_bitrate_addblock(vb);

		while (vorbis_bitrate_flushpacket(vd, &op)) {
			/* weld the packet into the bitstream */
			ogg_stream_packetin(os, &op);
			
			/* write out pages (if any) */
			do {
				if(ogg_stream_pageout(os, &og) == 0)
					break;

				zs->write(zs->cb_handle, og.header,
					og.header_len);
				zs->write(zs->cb_handle, og.body, 
					og.body_len);
			} while (!ogg_page_eos(&og));
		}
	}

	return T_OK;
}

/**** speex funcs ****/
/* TODO */
static int tz_init_speex(struct thinc_zsound *zs, va_list params)
{
	return T_ERROR;
}

static int tz_prepare_speex(struct thinc_zsound *zs)
{
	return T_ERROR;
}

static int tz_encode_speex(struct thinc_zsound *zs, char *buffer,
	unsigned int len)
{
	return T_ERROR;
}

static int tz_finish_speex(struct thinc_zsound *zs)
{
	return T_ERROR;
}

static void tz_end_speex(struct thinc_zsound *zs)
{
	free(zs->handle);
}
