/* 
 * doubly linked list implementation for thinc. derived from openbsd
 * circular queue code found at sys/queue.h which carries the following 
 * license and copyright:
 *
 * Copyright (c) 1991, 1993
 *      The Regents of the University of California.  All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. Neither the name of the University nor the names of its contributors
 *    may be used to endorse or promote products derived from this software
 *    without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE REGENTS AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE REGENTS OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 *
 *      @(#)queue.h     8.5 (Berkeley) 8/20/94
 */

/* main modification was to make the list let us remove elements without
 * having to pass a pointer to the head. this meant making the list head
 * look no different than the elements of the list. see comments below
 * for specific details */

/* set tabstop = 8 */

#ifndef _THINC_LIST_H_
#define _THINC_LIST_H_

/*
 * List definitions.
 */
/* bsd makes the list type specific but in our case we just want the list
 * to point to other list structs, that way the head is no different than
 * the nodes */
typedef struct thinc_list_s {
	struct thinc_list_s *next;
	struct thinc_list_s *prev;
#ifdef T_GETSTATS	/* baggage and more bagagge :( */
	unsigned int len;
	unsigned int max_len;
#endif	/* T_GETSTATS */
} thinc_list;

/*
 * List initializers 
 */
#define t_list_init(head) do {					\
	(head)->next = head;					\
	(head)->prev = head;					\
	t_list_len_init(head);					\
} while(0)

#ifdef T_GETSTATS
#define t_list_head_init(head) { t_list_end(&head), t_list_end(&head), 0, 0 }
#else
#define t_list_head_init(head) { t_list_end(&head), t_list_end(&head)}
#endif

#define t_list_alloc(head) do {					\
	head = malloc(sizeof(thinc_list));			\
	if(head)						\
		t_list_init(head);				\
} while(0)

/* 
 * If somebody uses head
 * after invalidated it we have
 * a segfault and lots of fun
 * guaranteed they are to have
 */
#define t_list_inv(head) 					\
	(head)->next = (head)->prev = NULL;

/*
 * List access methods
 */
/* these two macros were taken from linux kernel. nice way to get the
 * structure that encapsulates a list struct. hope they don't sue us :( */
#define t_offsetof(type, field) \
	((size_t) &((type*)0)->field)
#define t_container_of(ptr, type, field) \
	( (type*)( (char*)(ptr) - t_offsetof(type, field) ) )

#define t_list_entry(elm, type, field)	t_container_of(elm, type, field)
#define t_list_end(head)		(head)
#define t_list_next(elm)		((elm)->next)
#define t_list_prev(elm)		((elm)->prev)
#define t_list_next_entry(var, type, field)				\
	t_list_entry((var)->field.next, type, field)
#define t_list_prev_entry(var, type, field)				\
	t_list_entry((var)->field.prev, type, field)

#define t_list_first(head)		((head)->next)
#define t_list_last(head)		((head)->prev)
#define t_list_first_entry(head, type, field)				\
	t_list_entry(t_list_first(head), type, field)
#define t_list_last_entry(head, type, field)				\
	t_list_entry(t_list_last(head), type, field)

#define t_list_is_empty(head)						\
	(t_list_first(head) == t_list_end(head))

/* checks if list has only one element. kind of a hack */
#define t_list_one_el(head)						\
	( ( (head)->next != (head) ) && ( (head)->next->next == (head) ) )

/*
 * loop over the list. with the _fast macros is not safe to remove elements
 * inside the loop. note that for the final condition we "illegaly" 
 * derefence the memory around the list head. better that than requiring
 * caller to define and pass another variable (namely a ListPtr)
 */
#define t_list_for_each_entry_fast(var, type, head, field)		\
	for(	(var) = t_list_first_entry(head, type, field);		\
		&(var)->field != t_list_end(head);			\
		(var) = t_list_next_entry(var, type, field)		\
	)

#define t_list_for_each_entry(var, next, type, head, field)		\
	for(	(var) = t_list_first_entry(head, type, field),		\
		(next) = t_list_next_entry(var, type, field);		\
		&(var)->field != t_list_end(head);			\
		(var) = (next), (next) = t_list_next_entry(next, type, field)\
	   )

#define t_list_for_each_entry_cont_fast(var, type, head, field)		\
	for(	;							\
		&(var)->field != t_list_end(head);			\
		(var) = t_list_next_entry(var, type, field)		\
	)

/* loop using ListPtr's */
#define t_list_for_each(elm, next, head)				\
	for(	(elm) = t_list_first(head);				\
		(elm) != (head);					\
		(elm) = t_list_next(elm);				\
	   )

/*
 * List functions
 */
#define t_list_add(elm, p, n) do {					\
	(n)->prev = (elm);						\
	(elm)->next = (n);						\
	(elm)->prev = (p);						\
	(p)->next = (elm);						\
} while(0)

#define t_list_add_after(elm, p) do {					\
	(elm)->prev = (p);						\
	(elm)->next = (p)->next;					\
	(p)->next->prev = (elm);					\
	(p)->next = (elm);						\
} while(0)

#define t_list_add_before(elm, n) do {					\
	(elm)->next = (n);						\
	(elm)->prev = (n)->prev;					\
	(n)->prev->next = (elm);					\
	(n)->prev = (elm);						\
} while(0)

#define t_list_add_head(elm, head) do {					\
	t_list_add(elm, head, (head)->next);				\
	t_list_len_inc(head);						\
} while(0)

/* add entry at the tail of the list (queues). can't reuse list_add because
 * we want to be able to use this with empty lists */
#define t_list_add_tail(elm, head) do {					\
	((head)->prev)->next = (elm); 					\
	(elm)->prev = (head)->prev; 					\
	(elm)->next = (head); 						\
	(head)->prev = (elm); 						\
	t_list_len_inc(head);						\
} while(0)

#define t_list_del_el(elm) do {						\
	(elm)->next->prev = (elm)->prev;				\
	(elm)->prev->next = (elm)->next;				\
	t_list_init(elm);						\
} while(0)

/*
 * Manipulate whole lists
 */
/* puts list src at the end of list dst. assumes src has something. if
 * you don't know use t_list_move. also src is emptied.
 * the idea is kind of do add_tail(src->next, dst) and then 
 * add_tail(src->prev, dst) */
#define t_list_join(dst, src) do {					\
	(src)->next->prev = (dst)->prev;				\
	(dst)->prev->next = (src)->next;				\
	(src)->prev->next = (dst);						\
	(dst)->prev = (src)->prev;					\
	t_list_len_add(dst, t_list_len(src));				\
	t_list_init(src);						\
} while(0)

#define t_list_move(dst, src) do {					\
	if(!t_list_is_empty(src)) {					\
		t_list_join(dst, src);					\
	}								\
} while(0)

/* empties a list by repeatedly calling func for each entry in the list
 * and deleting it from the list. the function call will be 
 * func(handle, entry) */
#define t_list_empty(var, type, head, field, func, handle) do {		\
	type* __tmp;							\
	t_list_for_each_entry(var, __tmp, type, head, field) {		\
		t_list_del_el(&(var)->field);					\
		func(handle, var);					\
	}								\
	t_list_init(head);						\
} while(0)

#define t_list_empty_simple(head) do {					\
	thincListPtr __one, __two;					\
	t_list_for_each(__one, __two, head) {				\
		t_list_del_el(__one);					\
	}								\
	t_list_init(head);						\
} while(0)

#ifdef T_GETSTATS
#define t_list_len_init(list) \
	(list)->len = (list)->max_len = 0;

#define t_list_len_inc(list) do { 		\
	if(++(list)->len > (list)->max_len)	\
		(list)->max_len = (list)->len;	\
} while(0)

#define t_list_len_dec(list) --(list)->len

#define t_list_len_add(list, l) do {		\
	(list)->len += (l);			\
	if((list)->len > (list)->max_len)	\
		(list)->max_len = (list)->len;	\
} while(0)

#define t_list_len_sub(list, l) (list)->len -= (l)

#define t_list_len(list) (list)->len
#define t_list_maxlen(list) (list)->max_len

#define t_list_len_set(list, l) (list)->len = (l)
#define t_list_maxlen_set(list, l) (list)->max_len = (l)

#else /* T_GETSTATS */
#define t_list_len_init(list)

#define t_list_len_inc(list)
#define t_list_len_dec(list) 
#define t_list_len_add(list, l)
#define t_list_len_sub(list, l)

#define t_list_len(list)
#define t_list_maxlen(list)

#define t_list_len_set(list, l)
#define t_list_maxlen_set(list, l)

#endif	/* T_GETSTATS */

#endif	/* _THINC_LIST_H_ */
