/**************************************************************************
 *
 * Copyright 2002-2006 Columbia University
 * All Rights Reserved
 *
 **************************************************************************/
/*
 * The THINC protocol
 *
 * Authors: 
 * Ricardo A. Baratto - Network Computing Lab - Columbia University
 * Leo Kim - Network Computing Lab - Columbia University
 */

/* set tabstop = 8 */

/*
 * This is the THINC protocol definition. Aims to be small, efficient and 
 * extensible.
 * There are Server->Client messages and Client->Server messages. Each
 * message has a corresponding type that identifies it. Types from
 * s->c may overlap types from c->s.
 * Each message is prepended by a header, thincHeader, which currently 
 * carries the message's type. 
 *
 * Rules for adding new messages
 * -----------------------------
 *
 * For message types:
 * Include the name of the structure that your particular type represents.
 * The limit is 255
 *
 * For the messages:
 * Every message should have a clear description of what it does, it's
 * parameters, the data (if any) that follows it, and the size of its 
 * header. 
 * If any of your fields take only a set of constant values use #define's 
 * and put them below your message's definition.
 * Follow the naming conventions of the existing message types.
 * If your particular message has even further "subdivisions", include
 * the main message header definition here, and the rest in an additional
 * header file, which gets #include'd at the end of this file.
 * On Size: 
 * You should create a #define showing your message's header size.
 * Some architectures align structures to 32bit boundaries, 
 * so by rule all message structures should include enough padding to
 * account for this (char pad[blah] should do it). However, when sending
 * and receiving them, the actual size (indicated by the #define mentioned 
 * above) should be used. In addition, all the fields should explicitly tell 
 * the compiler their size using bitfields.
 * Don't use unsigned fields, java (one of our "clients") doesn't know about 
 * those.
 */
/* 
 * Current connection setup process:
 * (msg type, client state)
 * [Note: client state is used internally by the server to keep track of
 * the stage at which a client is]
 *
 * 1. Client connects and sends its string version. (T_CSTART, TCLIENT_INIT)
 * 2. If server "likes" the client's version, it sends back its version
 * along with its endian-ness information and any other information
 * relevant to the client. (T_SSTART, TCLIENT_INIT). 
 * If it doesn't "like" it, the server sends back an negative T_SACK message 
 * and closes the connection.
 **** START CHANGE
 * CHANGE: On version 0.1.1:
 * 3. If the client "likes" the server version, its endian-ness, and has
 * no problems with any additional information, it sends back an ack
 * telling the server to send the initial framebuffer (T_CACK, TCLIENT_START)
 * 4. Server sends initial framebuffer and cursor information. The cursor
 * information is optional, in case the server doesn't support hardware
 * cursors (T_FB_INIT, TCLIENT_FB)
 **** CHANGE: On version 0.1.2:
 * 3. If the client "likes" the server version, its endian-ness, and has
 * no problems with any additional infomartion, it sends back the user
 * authentication information. (T_CAUTH, TCLIENT_AUTH)
 * 4. Server authenticates the user (currently the user must have a valid
 * account on the system and be the owner of the session). If authentication
 * fails, the server sends back a negative T_SACK message and closes the
 * connection. If authentication succeeds, the server sends back the initial 
 * framebuffer and cursor information. The cursor information is optional, in 
 * case the server doesn't support hardware cursors (T_FB_INIT, TCLIENT_FB)
 **** END CHANGE
 * 5. If client is capable of handling this framebuffer and cursor, it sends 
 * back another ack telling the server to send the fb (T_CACK, TCLIENT_CONN)
 * 6. Server sends initial cursor and framebuffer to client 
 * (message its data part of T_FB_INIT, TCLIENT_CONN).
 */

/* TODO:
 * - For the video, the server should inform the client at connection time
 *   if it'll be negotiating video formats or not.
 * - Add negotiation of features in connection setup.
 * - Add padding to message's data
 * - Add maximum size to a message (with data included)
 */

#ifndef _THINC_PROTO_H_
#define _THINC_PROTO_H_

/**************************** Message header ******************************/
typedef struct thinc_header_s {
	int type:8;	/* 8 bits seems more than enough. If we have more
			   than 255 messages we're in trouble */
	int flags:8;	/* message dependent */
	int length:16;	/* length of the message. 
			   FIXME: This is only used sometimes. */
} thinc_header;

#define THINC_HEADERSIZE 4

/*************************************************************************
 *                          SERVER => CLIENT                             *
 *************************************************************************/

/**************************** Message types ******************************/

#define T_SSTART	1	/* thinc_sStart */
#define T_SPING		9	/* thinc_sPing */
/* 2 - 9 reserved for future use */
#define T_SACK		10	/* thinc_sAck */
/* Framebuffer painting messages */
#define T_FB_INIT	11	/* thinc_fbInit */
#define T_FB_RUPDATE	12	/* thinc_rawUpdate */
#define T_FB_COPY	13	/* thinc_fbCopy */
#define T_FB_SFILL	14	/* thinc_fbSFill */
#define T_FB_PFILL	15	/* thinc_fbPFill */
#define T_FB_GLYPH	16	/* thinc_fbGlyph */
#define T_FB_BILEVEL	17	/* thinc_fbBilevel */
/* - 39 reserved for framebuffer operations */
/* Cursor events */
#define T_CURSOR_CHANGE	40	/* thinc_cursorChange */
#define T_CURSOR_SHOWHIDE	41	/* thinc_cursorShowHide */
#define T_CURSOR_MOVE	42	/* thinc_cursorMove */
#define T_CURSOR_COLOR	43	/* thinc_cursorColor */

/* useful to keep in sync with things that depend on the protocol ids */
#define T_FB_FIRST_MSG	12
#define T_FB_LAST_MSG	17

/* 
 * current allocations:
 * resize: 50-60
 * video: 100-110
 */
/********************** helper structures ******************************/

/* description of a rectangular region */
typedef struct thinc_rectRegion_s {
	int x		:16;
	int y		:16;
	int width	:16;
	int height	:16;
} thinc_rectRegion;

#define T_RECTREGION_SIZE	8

/**************************** The messages *******************************/

/* T_SSTART
 * Description: Initial message sent when a new connection is received.
 * First 8 bits represent endian-ness of the server, T_BIG T_LITTLE. 
 * Next 24 bits represent the protocol version. XXX: Must discuss.
 * This is sent in response to a client T_CSTART message.
 * Server is free to reject a new connection based on this version string.
 * If the client decides to continue the connection, then all 
 * subsequent messages will be in the byte order specified by the byte_order
 * field (which should normally be the server's host machine byte order)
 * Size: 4 bytes
 * Data: NONE
 */
typedef struct thinc_sStart_s {
	int byte_order	:8;
	char version[3];
} thinc_sStart;

#define T_SSTART_SIZE 	4

/* possible byte-order values */
#define T_BIG 		0
#define T_LITTLE 	1

/* T_SPING
 * Description: kind of a keep alive. The message doesn't carry
 * anything beyond the header. The purpose is to try and keep the server to
 * client link from getting idle.
 * Size: 0
 * Data: NONE
 */

#define T_SPING_SIZE	0

/* T_SACK
 * Description: Used by the server to acknowledge certain messages. It's
 * a 1 or 0, representing the server's acceptance or denial to a particular 
 * client message.
 * Size: 1
 * Data: NONE
 */
typedef struct thinc_sAck_s {
	int ack		:8;
	char pad[3];
} thinc_sAck;

#define T_SACK_SIZE	1

/* T_FB_INIT
 * Description: Message sent when a new connection is made. It carries the 
 * information about the framebuffer and its cursor.
 * Size: 8 bytes
 * Data: Followed by cursorSize bytes representing the cursor and height 
 * lines of size width*bpp representing the contents of the framebuffer. 
 * However this data is only sent upon receipt of a T_CACK message
 */
typedef struct thinc_fbInit_s {
	/* fb information */
	int width	:16;	/* fb dimensions */
	int height	:16;
	int depth	:8;	/* color depth of fb */
	int bpp		:8;	/* how many bytes are used for each pixel */
	/* cursor information */
	int cursorSize	:16;	/* If positive, has the size of the cursor in 
				   bytes (if cursor is not argb, the source 
				   and the mask each are of size cursorSize/2.
				   otherwise, cursor is cursorSize bytes of 
				   pixel data)
				   If -1, no cursor handling will be
				   done by the client and the rest of the
				   information in this message should be 
				   disregarded */
	int cursorFG	:32;	/* cursor colors. if argb, ignore */
	int cursorBG	:32;
	int cursorWidth	:8;	/* cursor size in pixels */
	int cursorHeight :8;
	int cursorX	:16;	/* cursor position */
	int cursorY	:16;
	int cursorXHot	:8;	/* cursor's hot spot */
	int cursorYHot	:8;
} thinc_fbInit;

#define T_FB_INIT_SIZE	24

/* Support for compressed initial framebuffer:
 * When FB_INIT is sent, the following flag will be on in the flags field
 * of the header.
 * The framebuffer will be sent by compressing each scanline, and wrapping
 * it in a T_FB_ZDATA message which gives the client enough information
 * to uncompress it appropiately.
 */
#define T_FB_COMPRESSED	0x01	/* in flags field of the header */
#define T_FB_CURSOR_ARGB 0x02	/* the cursor is argb. cursorFG and cursorBG
				   should be ignored, and cursorSize is
				   the total data size in bytes of pixel
				   data of the cursor */

/*
 * T_FB_ZDATA
 * Description: Wrapper for a compressed update. It contains the compressed
 * and uncompressed size of the data following it.
 * Size: 4 bytes
 * Data: zsize bytes of compressed data
 */
typedef struct thinc_fbZData_s {
	int zsize	:32;	/* length of the compressed scanline */
	int usize	:32;	/* length of the uncompressed scanline */
} thinc_fbZData;

#define T_FB_ZDATA_SIZE	8

/* T_FB_RUPDATE
 * Description: Message sent for a raw rectangular frame buffer update. 
 * Carries the description of the location and size of the update.
 * Size: 8 bytes
 * Data: Followed by height lines of size width*bpp, representing the
 * new contents of the described frame buffer region
 */
typedef struct thinc_rawUpdate_s {
	int x		:16;	/* coords of top-left corner of the update */
	int y		:16;
	int width	:16;	/* dimensions of the update */
	int height	:16;
} thinc_rawUpdate;

#define T_FB_RUPDATE_SIZE 8

/* flags */				/* order */
#define T_FB_RUPDATE_COMPRESSED	0x01	/* 3 */
#define T_FB_RUPDATE_RESIZED	0x02	/* 1 */
#define T_FB_RUPDATE_CACHED	0x04	/* 2 */
#define T_FB_RUPDATE_ADDCACHE	0x08	/* 2 */

/* T_FB_COPY
 * Description: Tells the client to copy the area of size width*height, 
 * with top left corner sx,sy to dx,dy.
 * Size: 12 bytes
 * Data: NONE
 */
typedef struct thinc_fbCopy_s {
	int sx	:16;
	int sy	:16;
	int dx	:16;
	int dy	:16;
	int width	:16;
	int height	:16;
} thinc_fbCopy;

#define T_FB_COPY_SIZE	12

/* T_FB_SFILL	
 * Description: Solid Fill command. Tells the client to fill the area 
 * described by x,y,w,h with the pixel value pixel.
 * Size: 12 bytes
 * Data: NONE
 */
#if 0
typedef struct thinc_fbSFill_s {
	int x	:16;
	int y	:16;
	int width	:16;
	int height	:16;
	int pixel	:32;
} thinc_fbSFill;

#define T_FB_SFILL_SIZE 12
#else
/* FIXME: Change name from "pixel" to "color" */
typedef struct thinc_fbSFill_s {
	int pixel	:32;
	int numrects	:16;
	char pad[2];
	/* followed by numrects thinc_rectRegion structs describing each
	 * of the regions to be filled with this colr */
} thinc_fbSFill;

#define T_FB_SFILL_SIZE	6

#endif

/* T_FB_PFILL
 * Description: Pixmap Fill command: tile pixmap along rectangles. x,y
 * tells the client where's the origin of the tiling region. Note that
 * the rectangles may (and most of the time will) describe a subregion
 * of the tiling region, for example if only the lower left corner of
 * the region needs to be painted. This information is necessary for
 * proper alignment of the pixmap, ie. the client shouldn't just paint
 * the pixmap starting at the origin of each of the rectangles, clipping
 * and alignment needs to be done.
 * Some implementation notes:
 * -> The pixmap will probably have a maximum size, above which some
 * other type of update should be used.
 * -> If the region to be filled is smaller than the pixmap, another
 * type of message should be used.
 * Size: 10 bytes
 * Data: The pixmap (width*height*bpp bytes) followed by numrects
 * thinc_rectRegion structures describing each of the regions that should
 * be filled with this pixmap.
 */
typedef struct thinc_fbPFill_s {
	/* the pixmap's origin */
	int x		:16;
	int y		:16;
	/* the pixmap size */
	int width	:16;
	int height	:16;
	/* number of rectangles */
	int numrects	:16;
	char pad[2];
	/* followed by the pixmap and numrects thinc_rectRegion's (see below) */
} thinc_fbPFill;

#define T_FB_PFILL_SIZE	10

#define T_FB_PFILL_RESIZED	0x01
#define T_FB_PIXMAP_CACHED	0x02
#define T_FB_PIXMAP_ADDCACHE	0x04

/* T_FB_GLYPH
 * Description: Glyph Fill command. Fill the region described by (x,y)
 * (width, height) using the bitmap as a stipple to fill the region:
 * If there's a 1 on the bitmap the color specified in the message
 * should be applied. If there's a 0 no operation should be performed
 * on that pixel.
 * Size: 12 bytes
 * Data: The bitmap to use as stipple. Bitmap's size is: 
 * ceiling(width/8)*height
 * where 8 is the number of pixels that are packed on every byte of the
 * bitmap. There's a macro defined in the sample server and client 
 * implementation for this number and for the bitmap's size, given its 
 * width and height.
 */
typedef struct thinc_fbGlyph_s {
	/* the region to fill */
	int x		:16;
	int y		:16;
	int width	:16;
	int height	:16;
	/* the color to use */
	int color	:32;
	/* number of rectangles. if 0 just blast the glyph onto the fb*/
	int numrects	:16;
	char pad[2];
	/* followed by a bitmap (width * height)/8 bytes long */
	/* followed by numrects thinc_rectRegion */
} thinc_fbGlyph;

#define T_FB_GLYPH_SIZE 14

/* T_FB_BILEVEL
 * Description: Bilevel Fill command. Fill the region described by (x,y)
 * (width, height) using the bitmap as a stipple to fill the region:
 * If there's a 1 on the bitmap the fg color specified in the message
 * should be applied. If there's a 0, the bg color specified in the
 * message should be applied.
 * Size: 16 bytes
 * Data: The bitmap to use as stipple. Bitmap's size is: 
 * ceiling(width/8)*height
 * where 8 is the number of pixels that are packed on every byte of the
 * bitmap. There's a macro defined in the sample server and client 
 * implementation for this number and for the bitmap's size, given its 
 * width and height.
 */
typedef struct thinc_fbBilevel_s {
	/* the region to fill */
	int x		:16;
	int y		:16;
	int width	:16;
	int height	:16;
	/* the colors to use */
	int fg		:32;
	int bg		:32;
	/* number of rectangles. If 0, just blast the bilevel onto the fb */
	int numrects	:16;
	char pad[2];
	/* followed by a bitmap (width * height)/8 bytes long */
	/* followed by numrects thinc_rectRegion */
} thinc_fbBilevel;

#define T_FB_BILEVEL_SIZE 18

/* flags for glyph and bilevel */
#define T_FB_BITMAP_CACHED	0x01	/* 1 */
#define T_FB_BITMAP_ADDCACHE	0x02	/* 1 */
#define T_FB_BITMAP_RESIZED	0x04

/**************** Cursor messages ******************/

/* T_CURSOR_CHANGE
 * Description: The cursor shape needs to be changed to the shape
 * being sent as data to this message. If the flags contains T_CURSOR_ARGB
 * then the new cursor is an argb image. Otherwise it's a bitmap
 * Size: 4 bytes
 * Data: If it's not argb, a bitmap of size size corresponding
 * to the coursor's source and mask (each of size size/2). If it's argb, 
 * size bytes of pixel data.
 */
typedef struct thinc_cursorChange_s {
	int xhot	:8;
	int yhot	:8;
	int size	:16;
} thinc_cursorChange;

#define T_CURSOR_CHANGE_SIZE	4

/* flags */
#define T_CURSOR_ARGB		0x01

/* T_CURSOR_SHOWHIDE
 * Description: Client should show/hide its current cursor according to
 * the contents of showhide: T_SHOW, T_HIDE.
 * Cursor Hide is an implementation dependent operation (it is required).
 * The server may choose to support it or replace it with an equivalent
 * operation (like sending a transparent cursor), and the client is free
 * to implement it if its windowing system allows such an operation,
 * implement it some other way (such as having a transparent
 * cursor on hand and using it when a Hide command comes in), or
 * ignoring it. The reasons for this:
 * -> Some windowing systems (eg. X) don't support a cursor hiding operation.
 * -> The sample server is able to implement cursor hide as a normal operation
 * or as sending a transparent cursor. It may decide either way depending
 * on the knowledge it has of what its clients support.
 * Size: 1 byte
 * Data: None
 */
typedef struct thinc_cursorShowHide_s {
	int showhide	:8;
	char pad[3];
} thinc_cursorShowHide;

#define T_SHOW	1
#define T_HIDE 	0

#define T_CURSOR_SHOWHIDE_SIZE	1

/* T_CURSOR_MOVE
 * Description: Client should move the cursor to the specified position.
 * Size: 4 bytes
 * Data: None
 */
typedef struct thinc_cursorMove_s {
	int x		:16;
	int y		:16;
} thinc_cursorMove;

#define T_CURSOR_MOVE_SIZE 	4

/* T_CURSOR_COLOR 
 * Description: Change the cursor's fg and bg colors. Only for non-argb
 * cursors.
 * Size: 8 bytes
 * Data: None
 */
typedef struct thinc_cursorColor_s {
	int fg		:32;
	int bg		:32;
} thinc_cursorColor;

#define T_CURSOR_COLOR_SIZE	8


/*************************************************************************
 *                          CLIENT => SERVER                             *
 *************************************************************************/

/**************************** Message types ******************************/

#define T_CSTART		1	/* thinc_cStart */
#define T_CAUTH			2	/* thinc_cAuth */
/* 2 - 9 reserved for future use */
#define T_CACK			10	/* thinc_cAck */
#define T_EV_MOTION		11	/* thinc_evMotion */
#define T_EV_BUTTON		12	/* thinc_evButton */
#define T_EV_KEYB		13	/* thinc_evKeyb */

/**************************** The messages *******************************/

/* T_CSTART
 * Description: Initial message sent when a new connection is opened.
 * Represent the version of the protocol the client knows about. If for
 * any reason the server doesn't like the version, it can close the 
 * connection. In response to this message the server sends a T_START
 * message which contains, among other things, the server's version. If
 * the client doesn't like that version, and same "version-liking" 
 * arguments apply.
 * This message should be in network byte order (big endian), since at 
 * this point no endian-ness information has been exchanged. After the 
 * client receives the T_START message from the server, it should start
 * using the byte order specified by the server.
 * Size: 3 bytes
 * Data: NONE
 */
typedef struct thinc_cStart_s {
	char version[3];
	char pad[1];
} thinc_cStart;

#define T_CSTART_SIZE 3

/* T_CAUTH
 * Description: Has authentication information from the user. The message
 * itself contains the lenght of the username and password. The actual
 * strings follow the message. The strings do not end in '\0' (as normal
 * C strings do) and are not padded
 * Size: 4 bytes
 * Data: Two strings, representing the username and password. The length
 * of each string is passed in the actual message
 */
typedef struct thinc_cAuth_s {
	int user_len:16;
	int pass_len:16;
	/* followed by username (user_len bytes) and password (pass_len bytes).
	 * the strings do not end in \0 */
} thinc_cAuth;

#define T_CAUTH_SIZE	4
 
/* T_CACK 
 * Description: Used by the client to acknowledge certain messages. It's
 * just 8 bits set to either 1 or 0, representing the clients acceptance
 * or denial to a particular server message.
 * Size: 1
 * Data: NONE
 */
typedef struct thinc_cAck_s {
	int ack		:8;
	char pad[3];
} thinc_cAck;

#define T_CACK_SIZE	1

/* T_EV_MOTION
 * Description: A mouse motion event being reported by the client. Position
 * reported as x,y coordinates.
 * Size: 4
 * Data: NONE
 */
typedef struct thinc_evMotion_s {
	int x		:16;	/* the location of the mouse */
	int y		:16;
} thinc_evMotion;

#define T_EV_MOTION_SIZE 	4

/* T_EV_BUTTON
 * Description: A button was pressed/released. The 8 bits encode the
 * information needed:
 *        7      6 - 0
 * [ Up/Down  | Button # ]
 * The button # is encoded as a number between 0-127. Are there mice
 * with 128 buttons?? :)
 * Up/Down should use T_UP/T_DOWN (see below)
 * Size: 1 
 * Data: NONE
 */
typedef struct thinc_evButton_s {
	int buttonMask	:8;	/* info about button that triggered the event*/
	char pad[3];
} thinc_evButton;

#define T_UP	0
#define T_DOWN	1

#define T_EV_BUTTON_SIZE 	1

/* T_EV_KEYB
 * Description: A key was pressed/released (T_DOWN/T_UP)
 * XXX: Currently, the key is encoded as an X keysym. Some discussion needs
 * to happen on this, especially considering the translation from Java
 * codes to keysym.
 * Size: 5
 * Data: NONE
 */
typedef struct thinc_evKeyb_s {
	int key		:32;	/* the code for the key which generated the
				   event */
	int down	:8;	/* it's pressed (T_DOWN) or released (T_UP) */
	char pad[3];
} thinc_evKeyb;

#define T_EV_KEYB_SIZE 	5

/* additional definitions. If you need to include more files put them here */
#include "thincProtoApp.h"
#include "thincProtoVideo.h"
#include "thincProtoResize.h"
#include "thincProtoCache.h"

#endif
