/**************************************************************************
 *
 * Copyright 2002-2006 Columbia University
 * All Rights Reserved
 *
 **************************************************************************/
/*
 * The THINC protocol, version 0.2
 *
 * Authors: 
 * Ricardo A. Baratto - Network Computing Lab - Columbia University
 */

/* set tabstop = 8 */

/* 
 * NOTES
 * This is the THINC protocol spec, version 0.2. The goal with this version
 * is to make the connection stage of the protocol not suck so much. As such,
 * all the changes here refer to the handshake stage. Messages used after
 * a client is fully connected do not change from 0.1.
 *
 * The basic idea with the new handshake is to make it possible at connection
 * time for the client and server to agree on a set of parameters that will
 * be used for the session. Indeed, 0.1 cannot even do that!
 *
 * A lot of this is modeled on the ssh protocol. Maybe one day we will just
 * scrap a lot of this and use something like SIP. But for now, we're in the
 * nih world. And we're still a hack.
 */

/* HANDSHAKE
 * 1 Server: version string (no header)
 * - Client: version string (no header)
 * 2 Server: security caps
 * - Client: security caps (which are a subset of the server caps)
 * - Server: session sec caps (or notok) (basically an ack from server)
 * 3 if(security caps.ssl) ssl_handshake (started by client)
 * 4 if(security caps.auth) auth (started by client, server replies ok/notok.
 *   TODO: Add multiple auth attempts)
 * 5 client starts sending requests to server to negotiate
 *   aspects of the connection. If the server does not understand a request,
 *   it sends back a reply with "unknown message". It should try not to close 
 *   the connection but i guess it's ok to send back a NOTOK if the request
 *   is just too insulting for the server.
 *   a server reply may be:
 *   => info client requested
 *   => do not support, can't send info, etc.
 *   => unknown request.
 *   if a request triggers multiple message exchanges, each part of the
 *   conversation should be identified by a different request. we need this
 *   to simplify the implementation of the server.
 *   Requests have their own message id space, and all are wrapped by a
 *   generic "request' structure.
 *   e.g. 
 *   client: request fb info
 *   server: send fb info
 *   client: request cursor
 *   server: no cursor.
 * 6 client: done. no more requests, we're all done, let's start doing
 *   real work.
 *
 * at this point, handshake is finished and normal operation begins.
 *
 * NOTES
 * At any point, a reject (not ok) may be sent to terminate the connection.
 */

/* SPECS
 * - Version String: "THINC_[major].[minor] [additional info]\0"
 * => [major] and [minor] are ascii numbers
 * => [additional info]: optional, maybe useful for debugging purposes
 * => must be NULL terminated
 * - Caps: 32bit integer with bits set marking specific caps.
 */

#ifndef _THINC_PROTO2_H_
#define _THINC_PROTO2_H_

/**********************
 * common definitions *
 **********************/
#ifdef THINC_HEADER_PROTOCOL_VERSION
#undef THINC_HEADER_PROTOCOL_VERSION
#endif

#define THINC_HEADER_PROTOCOL_VERSION	"0.2"

#include "thincProto.h"

/*****************************
 * server -> client messages *
 *****************************/

/*** message types ***/
typedef enum {
	T_SERV_OK=1,		/* server positive ack */
	T_SERV_NOTOK,		/* server negative and fatal ack */
	T_SERV_SEC_CAPS,	/* server security caps */
	T_SESS_SEC_CAPS,	/* the session security caps */
	T_SERV_REPLY,		/* server reply to a client request */
	T_SERV_LAST		/* just a marker */
} thinc_serverMsgs;

/* list of security capabilities */
#define T_CAPS_SEC_ENC		0x0001	/* encryption */
#define T_CAPS_SEC_AUTH		0x0002	/* authentication */

/*** message definitions ***/

/* T_SERV_SEC_CAPS
 * Description: The security capabilities of the server. See above for
 * available capabilities
 * Size: 4
 * Data: None
 */
typedef struct thinc_serverSecCaps_s {
	int caps:32;	/* 1 bit for each capability */
} thinc_serverSecCaps;

#define T_SERV_SEC_CAPS_SIZE	4

/* T_SESS_SEC_CAPS
 * Description: The security capabilities for the session. 
 * Size: 4
 * Data: None
 */
typedef struct thinc_sessionSecCaps_s {
	int caps:32;	/* 1 bit for each capability */
} thinc_sessionSecCaps;

#define T_SESS_SEC_CAPS_SIZE	4

/* T_SERV_OK
 * Description: A good reply from the server.
 * Size: 0
 * Data: None
 */
/*** no need for struct, just the type in the header is enough ***/

#define T_SERV_OK_SIZE		0

/* T_SERV_NOTOK
 * Description: A negative ack from the server. After this is sent, the
 * connection should be closed.
 * Size: 4
 * Data: (Optional) reason for NOTOK, plaintext string
 */
typedef struct thinc_serverNotOk_s {
	int reason_len:32;	/* if 0, no reason */
	/* possibly followed by string describing reason for reject */
} thinc_serverNotOk;

#define T_SERV_NOTOK_SIZE	4

/*****************************
 * client -> server messages *
 *****************************/

/*** message types ***/
typedef enum {
	T_CLIENT_OK=1,		/* client ack */
	T_CLIENT_NOTOK,		/* client negative ack */
	T_CLIENT_SEC_CAPS,	/* client security caps */
	T_CLIENT_AUTH,		/* client auth info */
	T_CLIENT_REQ,		/* client handshake request */
	T_CLIENT_DONE,		/* client is done with handshake */
	T_CLIENT_LAST		/* just a marker */
} thinc_clientMsgs;

/* Hacks so we can live with v1 */
#undef T_CAUTH
#define T_CAUTH	T_CLIENT_AUTH

/*** message definitions ***/

/* T_CLIENT_SEC_CAPS
 * Description: The security capabilities of the client. See above for 
 * available capabilities
 * Size: 4
 * Data: None
 */
typedef struct thinc_clientSecCaps_s {
	int caps:32;	/* 1 bit for each capability */
} thinc_clientSecCaps;

#define T_CLIENT_SEC_CAPS_SIZE	4

/* T_CLIENT_NOTOK
 * Description: A negative reply from the client. After this is sent, the
 * connection should be closed.
 * Size: 4
 * Data: (Optional) reason for NOTOK, plaintext string
 */
typedef struct thinc_clientNotOk_s {
	int reason_len:32;	/* if 0, no reason */
	/* possibly followed by string describing reason for reject */
} thinc_clientNotOk;

#define T_CLIENT_NOTOK_SIZE	4

/* T_CLIENT_OK
 * Description: Positive reply from client.
 * Size: 0
 * Data: None
 */
/*** no need for struct, just the type in the header is enough ***/

#define T_CLIENT_OK_SIZE	0

/* T_CLIENT_DONE
 * Description: Tells server that the handshake is done
 * Size: 0
 * Data: None
 */
/*** no need for struct, just the type in the header is enough */

#define T_CLIENT_DONE_SIZE	0

/********************************
 * session features negotiation *
 ********************************/

/* T_CLIENT_REQ
 * Description: A handshake request from the client. Identifies and precedes 
 * the specific request being made. You should think of it not as the client
 * asking for something from the server, but as the beginning of a 
 * negotiation with the server, as requests can carry client information too,
 * or they may just be used by the client to inform the server of "something"
 * (in which case the server will just reply "OK")
 * Size: 4
 * Data: None
 */
typedef struct thinc_clientReq_s {
	int request	:16;
	int unused	:16;
} thinc_clientReq;

#define T_CLIENT_REQ_SIZE	4

#define t_req_head	thinc_clientReq	req
#define t_req_size(req)	(T_CLIENT_REQ_SIZE + req##_SIZE)
#define t_req_type(r)	(r)->req.request
#define t_req_set_type(r, type) do {	\
	t_req_type(r) = htons(type);	\
	(r)->req.unused = 0;		\
} while(0)
#define t_req_read_type(r)	ntohs(t_req_type(r))

/* T_SERV_REPLY
 * Description: A reply to a client request. The id may be either one of
 * the thinc_serverReplies below or the id of the client's request to 
 * which we're replying. It may wrap around an additional reply message
 * containing data needed for the reply.
 * IMPORTANT: Use the header.length field so that client can read the whole
 * reply easily without having to figure out which reply you sent first
 * Size: 4
 * Data: None
 */
typedef struct thinc_serverReply_s {
	int reply:16;		/* id of the reply */
	int unused:16;
} thinc_serverReply;

#define T_SERV_REPLY_SIZE	4

#define t_reply_head	thinc_serverReply reply
#define t_reply_size(reply)	(T_SERV_REPLY_SIZE + reply##_SIZE)
#define t_reply_type(r)	(r)->reply.reply
#define t_reply_set_type(r, type) do {	\
	t_reply_type(r) = htons(type);	\
	(r)->reply.unused = 0;		\
} while(0)
#define t_reply_read_type(r)	ntohs((r)->reply.reply)

/* basic replies */
typedef enum {
	T_REPLY_OK,			/* positive reply */
	T_REPLY_NOTOK,			/* negative (non-fatal) reply */
	T_REPLY_UNKNOWN,		/* unknown request */
	T_REPLY_LAST			/* marker for requests below */
} thinc_serverReplies;

/* list of requests */
/* TODO: 
 * - Resize: so client can tell server it's current size (probably before
 * it sends the initial fb data)
 * - Clip: for scrolling. useful on pda client.
 * - Appsharing: more complicated. server needs to tell client which windows
 *   can be shared, and client must tell server which ones it wants.
 */
typedef enum {
	T_REQ_FBINFO=T_REPLY_LAST	/* basic framebuffer info */
	, T_REQ_CURSOR			/* cursor information and data */
	, T_REQ_FBDATA			/* contents of framebuffer */
	, T_REQ_ENCODER			/* how is image data encoded */
	, T_REQ_CACHESZ			/* size of various caches */
	, T_REQ_VIDEO			/* server supports video? */
	, T_REQ_NOVIDEO			/* client telling server that it 
					   can't support video (see below) */
	, T_REQ_VIDEO_SERV_FMTS		/* video formats supported by server */
	, T_REQ_VIDEO_CLIENT_FMTS	/* client informs server of its video
					   formats */
	, T_REQ_KEEPALIVE		/* client supports or not keepalives
					   does server support them? */
	, T_REQ_APP_SHARING		/* does the server support application
					   sharing? */
	, T_REQ_APP_LIST		/* list of available applications */
	, T_REQ_APP_GET			/* request a particular application.
					   client may send this message
					   multiple times */
} thinc_clientReqs;

/* request and reply messages. 
 * for requests, first member _must_ be thinc_clientReq. for replies
 * first member _must_ be thinc_serverReply.
 * use t_request_head and t_reply_head for this
 * BIG ASSUMPTION: compiler won't add pad after this struct since it's
 * aligned to 4-byte boundary.
 *
 * The *_SIZE #defines do not include the size of the reply/request header.
 * use t_reply_size() and t_req_size() to compute the total size if 
 * needed
 */

/* simple replies */
typedef struct thinc_replyOK_s {
	t_reply_head;
} thinc_replyOK;

#define T_REPLY_OK_SIZE		0

typedef struct thinc_replyNotOK_s {
	t_reply_head;
} thinc_replyNotOK;

#define T_REPLY_NOTOK_SIZE	0

typedef struct thinc_replyUnknown_s {
	t_reply_head;
} thinc_replyUnknown;

#define T_REPLY_UNKNOWN_SIZE		0

/****** FBINFO ******************/
typedef struct thinc_reqFBInfo_s {
	t_req_head;
} thinc_reqFBInfo;

#define T_REQ_FBINFO_SIZE	0

typedef struct thinc_replyFBInfo_s {
	t_reply_head;
	int byteOrder	:1;		/* T_BIG | T_LITTLE (thincProto.h) */
	int mbz		:15;		/* reserved for future use */
	int depth 	:8;
	int bpp		:8;		/* bits per pixel */
	int width	:16;
	int height	:16;
} thinc_replyFBInfo;

#define T_REPLY_FBINFO_SIZE	8

/****** CURSOR ******************/
typedef struct thinc_reqCursor_s {
	t_req_head;
} thinc_reqCursor;

#define T_REQ_CURSOR_SIZE	0

typedef struct thinc_replyCursor_s {
	t_reply_head;
	int flags	:16;	/* for now just: is it argb or not */
	int size	:16;	/* size of cursor data (bytes) */
	int width	:8;	/* dimensions */
	int height	:8;
	int xhot	:8;	/* hot spot */
	int yhot	:8;
	int x		:16;	/* position */
	int y		:16;
	int fg		:32;	/* colors. ignore if ARGB */
	int bg		:32;
	/* followed by size bytes representing the cursor */
} thinc_replyCursor;

#define T_REPLY_CURSOR_SIZE	20

/* flags */
/* reuse from cursor change message
#define T_CURSOR_ARGB	0x01
 */

/****** ENCODER *****************/
typedef struct thinc_reqEncoder_s {
	t_req_head;
} thinc_reqEncoder;

#define T_REQ_ENCODER_SIZE	0

typedef struct thinc_replyEncoder_s {
	t_reply_head;
	int which	:2;	/* zlib, png, none. see below */
	int reserved	:30;	/* reserved for future use */
} thinc_replyEncoder;

#define T_REPLY_ENCODER_SIZE	4

#define T_ENCODER_NONE		0x00
#define T_ENCODER_PNG		0x01
#define T_ENCODER_ZLIB		0x02

/****** FBDATA *******************/
typedef struct thinc_reqFBData_s {
	t_req_head;
} thinc_reqFBData;

#define T_REQ_FBDATA_SIZE	0

typedef struct thinc_replyFBData_s {
	t_reply_head;
	int flags	:32;
	/* if compressed, followed by fbZData. if not, followed by
	 * pixel data */
} thinc_replyFBData;

#define T_REPLY_FBDATA_SIZE	4

/* (already in thincProto.h)
#define T_FB_COMPRESSED		0x01
 */

/****** CACHESIZE ***************/
/* TODO: Should the client tell the server which caches it supports? */
typedef struct thinc_reqCacheSz_s {
	t_req_head;
} thinc_reqCacheSz;

#define T_REQ_CACHESZ_SIZE	0

typedef struct thinc_replyCacheSz_s {
	t_reply_head;
	/* all cache sizes are in bits */
	int		img:8;
	int		bit:8;
	int		pix:8;
	int		reserved:8; 
} thinc_replyCacheSz;

#define T_REPLY_CACHESZ_SIZE	4

/****** VIDEO *******************/
/* server replies ok or notok */
typedef struct thinc_reqVideo_s {
	t_req_head;
} thinc_reqVideo;

#define T_REQ_VIDEO_SIZE	0

/* server doesn't reply (unless he or she feels insulted) */
typedef struct thinc_reqNoVideo_s {
	t_req_head;
} thinc_reqNoVideo;

#define T_REQ_NOVIDEO_SIZE	0

typedef struct thinc_reqVideoServFmts_s {
	t_req_head;
} thinc_reqVideoServFmts;

#define T_REQ_VIDEO_SERV_FMTS_SIZE	0

typedef struct thinc_replyVideoServFmts_s {
	t_reply_head;
	thinc_vidSInit	fmts;	/* from thincProtoVideo.h */
} thinc_replyVideoServFmts;

#define T_REPLY_VIDEO_SERV_FMTS_SIZE	T_VIDEO_SINIT_SIZE

/* server replies ok or notok */
typedef struct thinc_reqVideoClientFmts_s {
	t_req_head;
	thinc_vidCInit	fmts;	/* from thincProtoVideo.h */
} thinc_reqVideoClientFmts;

#define T_REQ_VIDEO_CLIENT_FMTS_SIZE	T_VIDEO_CINIT_SIZE

/****** APPLICATION SHARING ****/
/* T_REQ_APP_SHARING
 * server replies ok or notok. If server replies ok (see thincProtoApp.h) each
 * thinc_header sent will be followed by a thinc_appHeader */
typedef struct thinc_reqAppSharing_s {
	t_req_head;
} thinc_reqAppSharing;

#define T_REQ_APP_SHARING_SIZE	0

/* T_REQ_APPLIST
 * client asks the server which applications can be shared. The idea is to
 * let the user choose which ones it wants.
 */
typedef struct thinc_reqAppList_s {
	t_req_head;
} thinc_reqAppList;

#define T_REQ_APP_LIST_SIZE	0

/* how many apps we're sending, followed by count thinc_appDesc structs */
typedef struct thinc_replyAppList_s {
	t_reply_head;
	int count		:8;
	int mbz			:8;
} thinc_replyAppList;

#define T_REPLY_APP_LIST_SIZE	2

/* T_REQ_APPLIST
 * ask the server for a particular app (from the list sent by APP_LIST). 
 * The server may reply notok... 
 */
typedef struct thinc_reqAppGet_s {
	t_req_head;
	int app_id		:8;
	int mbz			:8;
} thinc_reqAppGet;

#define T_REQ_APP_GET_SIZE	2

typedef struct thinc_replyAppGet_s {
	t_reply_head;
} thinc_replyAppGet;

#define T_REPLY_APP_GET_SIZE	0

/* this union allows the server to read the request in one swoop (using
 * header.length field for size). */
typedef union thinc_clientReqGen_s {
	t_req_head;
	thinc_reqFBInfo 	fbinfo;
	thinc_reqCursor 	cursor;
	thinc_reqEncoder	encoder;
	thinc_reqFBData		fbdata;
	thinc_reqCacheSz	cachesz;
	thinc_reqVideo		video;
	thinc_reqNoVideo	novideo;
	thinc_reqVideoServFmts	video_serv_fmts;
	thinc_reqVideoClientFmts video_client_fmts;
	thinc_reqAppSharing	app_sharing;
	thinc_reqAppList	app_list;
	thinc_reqAppGet		app_get;
} thinc_reqGen;

/* this union allows the client to read the request in one swoop (using
 * header.length field for size). */
typedef union thinc_clientReplyGen_s {
	t_reply_head;
	thinc_replyOK		ok;
	thinc_replyNotOK	notok;
	thinc_replyUnknown	unknown;
	thinc_replyFBInfo 	fbinfo;
	thinc_replyCursor 	cursor;
	thinc_replyEncoder	encoder;
	thinc_replyFBData	fbdata;
	thinc_replyCacheSz	cachesz;
	thinc_replyVideoServFmts video_serv_fmts;
	thinc_replyAppList	app_list;
	thinc_replyAppGet	app_get;
} thinc_replyGen;

/*******
 * end *
 *******/
#endif	/* _THINC_PROTO2_H_ */
