/**************************************************************************
 *
 * Copyright 2002-2006 Columbia University
 * All Rights Reserved
 *
 **************************************************************************/
/* 
 * Video protocol for THINC
 *
 * Authors: 
 * Ricardo A. Baratto - Network Computing Lab - Columbia University
 */
/* set tabstop = 8 */

#ifndef _THINC_PROTO_VIDEO_H_
#define _THINC_PROTO_VIDEO_H_

/* Bird's-eye view:
 * Upon initialization, the server and client negotiate video formats.
 * If a consensus is reached, video support is enabled.
 * When a video needs to be played, the server sends a message to the client
 * requesting a video channel. The request contains details about the
 * video such as position on the screen, size, and format. If the client
 * is able to support it, it sends back an acknowledgement to the server.
 * Subsequently, the server streams the video to the client. The way
 * the video is sent is dependent on the format.
 */

/* server -> client */
#define T_VIDEO_SINIT	100	/* thinc_vidSInit */
#define T_VIDEO_START	110	/* thinc_vidStart */
#define T_VIDEO_NEXT	111	/* thinc_vidNext */
#define T_VIDEO_END	112	/* thinc_vidEnd */
#define T_VIDEO_MOVE	113	/* thinc_vidMove */
#define T_VIDEO_SCALE	114	/* thinc_vidScale */
#define T_VIDEO_RESIZE	115	/* thinc_vidResize */

/* T_VIDEO_SINIT
 * Description: Server video inialization msg. Accompanied by formats 
 * supported by the server. Contains number of video formats.
 * Size: 4 bytes
 * Data: The video formats
 */
typedef struct thinc_vidSInit_s {
	int numformats	:32;
	/* followed by array of numformats thinc_vidFormat (below) */
} thinc_vidSInit;

#define T_VIDEO_SINIT_SIZE	4

/* For now we only use FOURCC format descriptions, which both the server
 * and client should be able to understand. In the future, a more thorough
 * description of the format may be necessary */
typedef struct thinc_vidFormat_s {
	int id		:32;	/* FOURCC id of this format */
} thinc_vidFormat;

#define T_VIDEO_FORMAT_SIZE	4

/* some format identifiers */
#define T_FOURCC_YV12 0x32315659  /* {'2','1','V','Y'} */
#define T_FOURCC_YUY2 0x32595559  /* {'2','Y','U','Y'} */
#define T_FOURCC_UYVY 0x59565955  /* {'Y','V','U','Y'} */

/* T_VIDEO_START
 * Description: Asks the client to start playing video which will come in 
 * the format specified in the message. It contains information about the 
 * video, such as format (which must be one of the agreed formats), 
 * dimensions, position, etc. It also defines an ID for this video stream.
 * The flags field in the header will have additional information about the
 * video stream, e.g. if it will be sent compressed.
 * Size: 20 bytes
 * Data: None
 */
typedef struct thinc_vidStart_s {
	int id		:32;	/* id for this video stream */
	/* format id */
	int fmt_id	:32;	/* format. Only send id, which should have
				   been set on initialization */
	/* position */
	int x		:16;
	int y		:16;
	/* dimensions */
	int width 	:16;
	int height 	:16;
	/* for client scaling */
	int dst_width	:16;
	int dst_height  :16;
	/* followed by struct with format specific info (possible list
	 * below) */
	/*
	struct thinc_vidInfoYV12 yv12;
	struct thinc_vidInfoYUY2 yuy2;
	struct thinc_vidInfoUYVY uyvy;
	*/
} thinc_vidStart;

#define T_VIDEO_COMPRESSED	0x01

#define T_VIDEO_START_SIZE	20

/* format specific messages */
typedef struct thinc_vidInfoYV12 {
	int yuv_yoffset:32;
	int yuv_uoffset:32;
	int yuv_voffset:32;
	int yuv_ypitch:32;
	int yuv_upitch:32;
	int yuv_vpitch:32;
} thinc_vidInfoYV12;

#define T_VIDEO_INFO_YV12_SIZE	24

typedef struct thinc_vidInfoYUY2 {
	int pitch:32;
} thinc_vidInfoYUY2;

#define T_VIDEO_INFO_YUY2_SIZE	4

typedef struct thinc_vidInfoUYVY {
	int pitch:32;
} thinc_vidInfoUYVY;

#define T_VIDEO_INFO_UYVY_SIZE	4

/* T_VIDEO_NEXT
 * Description: Data that should be used next to continue playing video
 * Size: 12 bytes
 * Data: The video data
 */
typedef struct thinc_vidNext_s {
	int id		:32;	/* stream id */
	int size	:32;	/* size of video data */
	int timestamp	:32;	/* timestamp for synchronization */
	/* followed by video data */
} thinc_vidNext;

#define T_VIDEO_NEXT_SIZE	12

/* T_VIDEO_END
 * Description: Video for stream id has finished. Client should clean things
 * up.
 * Size: 4 bytes
 * Data: None
 */
typedef struct thinc_vidEnd_s {
	int id		:32;
} thinc_vidEnd;

#define T_VIDEO_END_SIZE	4

/* T_VIDEO_MOVE
 * Description: Move the position of the video stream id to the new
 * coordinates
 * Size: 8 bytes
 * Data: None
 */
typedef struct thinc_vidMove_s {
	int id		:32;
	int x		:16;
	int y		:16;
} thinc_vidMove;

#define T_VIDEO_MOVE_SIZE	8

/* T_VIDEO_SCALE
 * Description: Change destination dimensions of the video stream id
 * Size: 8 bytes
 * Data: None
 */
typedef struct thinc_vidScale_s {
	int id		:32;
	int dst_width	:16;
	int dst_height	:16;
} thinc_vidScale;

#define T_VIDEO_SCALE_SIZE	8

/* T_VIDEO_RESIZE
 * Description: Change the source size of the video. Source size represents
 * how big the frames being sent by the server to the client are. Used when
 * client's window is smaller than original video.
 * Size: 8 bytes
 * Data: None
 */
typedef struct thinc_vidResize_s {
	int id		:32;
	int width	:16;
	int height	:16;
	/* followed by struct with format specific info (possible list
	 * below) */
	/*
	struct thinc_vidInfoYV12 yv12;
	*/
} thinc_vidResize;

#define T_VIDEO_RESIZE_SIZE	8

/************** client -> server *******************/
#define T_VIDEO_CINIT	100	/* thinc_vidCInit */
#define T_VIDEO_STARTOK	110	/* thinc_vidStartOK */

/* T_VIDEO_CINIT
 * Client initialization msg. Sent in response to a VIDEO_SINIT message.
 * It contains the subset of formats which the client can support. If the 
 * list is empty, video initialization fails */
typedef struct thinc_vidCInit_s {
	int numformats	:32;
	/* followed by array of numformats thinc_vidFormat (above) */
} thinc_vidCInit;

#define T_VIDEO_CINIT_SIZE	4

/* T_VIDEO_STARTOK
 * Response to T_VIDEO_START. If ok is T_VIDEO_OK, client is able to play 
 * the video. Otherwise, video can't be played and server should refrain 
 * from sending video data */
typedef struct thinc_vidStartOK_s {
	int		id:32;	/* video stream for which client is replying */
	int 		ok:8;	/* T_VIDEO_OK/T_VIDEO_NOTOK */
	char pad[3];
} thinc_vidStartOK;

#define T_VIDEO_OK	1
#define T_VIDEO_NOTOK	0

#define T_VIDEO_STARTOK_SIZE	5

#endif
