#!/usr/bin/python
#
# script to parse output from instrumented X server
#
# Authors: 
# Ricardo Baratto - Network Computing Lab - Columbia University
#
# Released to the public domain. You may use this
# code any way you wish, private, educational, or commercial.  It's free.
####
# set ts = 4
####

import os, string, sys

verbose = False

def msg(msg):
	if verbose:
		print msg

def parseOptions():
	from optparse import OptionParser
	opt_parser = OptionParser()
	opt_parser.add_option(
		"-v",
		"--verbose",
		action="store_true",	# flag is turned on
		dest="verbose",
		help="print what's going on"
	)
	opt_parser.add_option(
		"-A",
		"--aggregates",
		action="store_true",
		dest="aggs",
		help="compute some aggregate values from the data"
	)
	opt_parser.add_option(
		"-t",
		"--timeline",
		action="store_true",
		dest="timeline",
		help = "generate a pretty timeline of the data. by default, the " \
			"script will try to analyze the data to compute latency and " \
			"stuff. this option disables that behavior"
	)
	opt_parser.add_option(
		"-c",
		"--cpuspeed",
		action="store",
		dest="cpuspeed",
		type="float",
		help="the speed of the cpu where the measurements where taken. " \
			"If not specified, i will parse /proc/cpuinfo. It should be " \
			"specified look like 449.041 (in MHz)"
	)
	return opt_parser.parse_args()

def file_iter(file):
	"""
	a file iterator. pass a file and it'll return a line at a time
	"""
	for line in open(file):
		yield line.rstrip()

def evsBegin(line):
	if line.find("EVS") == -1:
		return False
	msg("found EVS line '%s'" % line)
	return True

def evsGet(line, ev_dict):
	### It adds an event pair to dict. returns False if it can't parse the line
	try:
		num, name = line.split(" ")
		name = name.replace("TI_", "").replace("_", " ")
		msg("num " + num + " name " + name)
		ev_dict[int(num)] = name
		return True
	except ValueError:
		return False

def evsEnd(line):
	if line.find("EVS END") == -1:
		return False
	msg("done with evs '%s'" % line)
	return True

def procsBegin(line):
	if line.find("PROCS") == -1:
		return False
	msg("found PROCS line '%s'" % line)
	return True

def procsGet(line, proc_dict):
	### It adds a proc pair to dict. returns False if it can't parse the line
	try:
		name, num = line.split(" ")
		name = name.replace("Proc", "")
		msg("num " + num + " name " + name)
		proc_dict[int(num)] = name
		return True
	except ValueError:
		return False

def procsEnd(line):
	if line.find("PROCS END") == -1:
		return False
	msg("done with procs '%s'" % line)
	return True

def instrStart(line):
	"""
	returns how many events we have or -1 if it can't parse the line
	"""
	if line.find("EVENTS") == -1:
		return -1
	else:
		list = line.split(" ")
		return int(list[1])

def instrEnd(line):
	if line.find("EVENTS END") == -1:
		return False
	else:
		return True

def parseLine(line, evs, procs):
	"""
	returns dict with parsed items. dict has following keys:
	id	- int
	event - string
	request - string
	time - float
	"""
	try:
		id, req, ev, thigh, tlow = line.split(" ")
	except ValueError:
		msg("can't parse your line. raising an exception. either " \
			"someone upstairs catches it or you're doomed")
		## pass on the exception to caller after printing a very useful
		## message
		raise ValueError

	## report just cycles
#	time = convertTime(long(tlow), long(thigh))
	## report microseconds
	cycles = convertTime(long(tlow), long(thigh))
	time = cyclesToUSec(cycles, cpuspeed)
	id = int(id)
	try:
		event = evs[int(ev)]
	except KeyError:
		event = "UNKNOWN " + ev
	try:
		request = procs[int(req)]
	except KeyError:
		request = "UNKNOWN " + req

	dict = {"id":id, "event":event, "request":request, "time":time}
	return dict

def isInputLine(string):
	if string.find("INPUT") == -1:
		return False
	else:
		return True

# - need 4 INPUTS. then when we get the next INPUT we're finished
def nextPage(iter, line, evs, procs):
	"""
	page is a dict with the following keys:
		start time
		end   time
	line is the current line. we need it because as this method is repeatedly
	called, it ends execution when it finds a line it shouldn't process
	(which is then passed to the next time it's called).
	iter provides line from a file.
	"""
	page = {}
	try:
		dict = parseLine(line, evs, procs)
	except ValueError:
		# can't parse this line. maybe we're at the end?
		return (None, None)
	# first line should be INPUT or something is off
	if not isInputLine(dict["event"]):
		## error
		print "first line in page should be INPUT event"
		raise IndexError ## random exception :)

	page["start"] = dict["time"]
	input_count = 1
	ev_count = 1
	curr_time = 0.0

	# process the rest of the lines until we find the fifth INPUT line
	for line in iter:
		try:
			dict = parseLine(line, evs, procs)
		except ValueError:
			# can't parse this line. maybe we're at the end?
			return (None, None)
		if isInputLine(dict["event"]):
			input_count += 1
#			msg("have %d inputs" % input_count)
			if input_count == 7:
				page["end"] = curr_time
				page["ev_count"] = ev_count
#				msg("we're done with this page")
				# we'll have to reparse this line. that's ok 
				return (page, line) 
		else:
			# just a normal page
			ev_count += 1
			curr_time = dict["time"]
	else:
		# let's be safe against crappy input
		print "Unexpected EOF. somebody sucks"
		return (None, None)

####### util functions
def convertTime(tlow, thigh):
	return (thigh << 32) + tlow

### assume speed is in MHz
def cyclesToSec(cycles, speed):
	return cycles/(speed*1000*1000)

def cyclesToUSec(cycles, speed):
	return cycles/speed

def secToCycles(cycles, speed):
	return cycles*speed

def getCpuspeed():
	for line in open("/proc/cpuinfo"):
		if line.find("cpu MHz") == -1:
			continue

		str, mhz = line.split(":")
		break

#	return float(mhz)*1024*1024;
	return float(mhz)

def getTabs(howmany):
	s = ""
	for i in range(0, howmany):
		s += "   "

	return s

def timeline(iter, evs, procs):
	prev_time = 0.0
	line = iter.next()
	count = instrStart(line)
	if count == -1:
		print "Your file sucks. can't parse EVENTS line '%s'" % line
		sys.exit(1)

	print "Have %d events for you, ready?" % count
	print "%-5s%-20s%-32s%-20s%-10s" % ("cl", "event", "request", "time", \
		"delta")

	if options.aggs:
		aggs = {}
		prev = 0.0

	for line in iter:
		try:
			dict = parseLine(line, evs, procs)
		except ValueError:
			# not a good line, we're probably finished with the events
			break

		count -= 1

		print "%-5s%-20s%-32s%-20.2f%-10.2f" % (dict["id"], dict["event"], \
			dict["request"], dict["time"], dict["time"]-prev_time)
		prev_time = dict["time"]

		if not options.aggs:
			continue
		
		# compute aggregate times
		if dict["event"].find("INPUT") != -1:
			continue
		elif dict["event"].find("START") != -1:
			if prev:
				try: 
					aggs["wait"] += dict["time"]-prev
				except KeyError:
					aggs["wait"] = dict["time"]-prev
			prev = dict["time"]
		elif dict["event"].find("END") != -1:
			try:
				aggs[dict["request"]] += dict["time"]-prev
			except KeyError:
				aggs[dict["request"]] = dict["time"]-prev
			prev = dict["time"]

	else:	## in case we don't do a break above
		print "unexpected EOF. your file sucks!"
		sys.exit(1)

	if count:
		print "didn't process all events. %d missing!" % count
		sys.exit(1)

	if instrEnd(line):
		msg("happy, happy, happy")
		print "au revoir!"
	else:
		msg("sad, sad, sad")
		print "you suck!"

	if options.aggs:
		print "aggregates:"
		for req in aggs.keys():
			print "%s: %.2fms (%.2fus)" % (req, aggs[req]/1000, aggs[req])

def analyze(iter, evs, procs):
	pages = []
	page_curr = {}
	page_count = 0

	line = iter.next()
	count = instrStart(line)
	if count == -1:
		print "I am sad. I can't parse EVENTS line '%s'" % line
		sys.exit(1)

	print "Have %d events for you. let's see what that's supposed to " \
		"mean..." % count
	
	line = iter.next()
	while True:
		try:
			(page_curr, line) = nextPage(iter, line, evs, procs)
		except IndexError: 
			# random exception raised by nextPage on error :)
			print "YOU SUCK"
			sys.exit(1)
		if not line:
			## we're done
			msg("dat's it. got %d pages" % page_count)
			break
		pages.append(page_curr)
		page_count += 1
		msg("got page #%d. start %.2f, end %.2f, delta %.2f count %d" % \
			(page_count, page_curr["start"], page_curr["end"], \
			page_curr["end"] - page_curr["start"], page_curr["ev_count"]))

	latency = getLatency(pages)
	latency_s = latency/1000000
	print "Pages: %d" % page_count
	print "Latency: %.2fs (%.2fus)" % (latency_s, latency)
	print "Per-Page: %.2fs (%.2fus)" % (latency_s/page_count, \
		latency/page_count)

def getLatency(pages):
	"""
	pages is list of dictionaries
	"""
	latency = 0.0
	count = 1
	msg("page  %-20s%-20s%-16s%-8s%-8s" % \
		("start", "end", "time", "time (s)", "ev count"))
	for page in pages:
		start = page["start"]
		end = page["end"]
		page_time = end-start
		latency += page_time
		msg("[%02d]: %-20.2f%-20.2f%-16.2f%-8.2f%-8d" % \
			(count, page["start"], page["end"], page_time, page_time/1000000, \
			page["ev_count"]) \
		)
		count += 1
	
	return latency
	
if __name__ == '__main__':
	(options, args) = parseOptions()

	try:
		fn = args[0]
	except IndexError:
		print "error: need to pass a filename boy"
		sys.exit(1)

	verbose = options.verbose

	evs = {}
	procs = {}
	if options.cpuspeed:
		cpuspeed = options.cpuspeed
	else:
		cpuspeed = getCpuspeed()
	print "your cpu has speed %.3fMHz" % cpuspeed

	msg("filename is '%s', here we go!" % fn)

	## parse EVS definitions
	iter = file_iter(fn)
	line = iter.next()
	if not evsBegin(line):
		print "couldn't find EVS. your file sucks!"
		sys.exit(1)

	for line in iter:
		if not evsGet(line, evs):
			break

	if not evsEnd(line):
		print "your file sucks. can't find EVS END"
		sys.exit(1)

	## parse PROCS definitions
	line = iter.next()
	if not procsBegin(line):
		print "couldn't find PROCS. your file sucks!"
		sys.exit(1)

	for line in iter:
		if not procsGet(line, procs):
			break

	if not procsEnd(line):
		print "your file sucks. can't find PROCS END"
		sys.exit(1)

	## now we're ready for some fun
	if options.timeline:
		timeline(iter, evs, procs)
	else:
		analyze(iter, evs, procs)

	sys.exit(0)
