/*
 * Program to display a RAW update on a X window
 *
 * Released to the public domain. You may use this
 * code any way you wish, private, educational, or commercial.  It's free.
 *
 * Authors: 
 * Ricardo A. Baratto - Network Computing Lab - Columbia University
 */

/* 
 * input files should have header:
 * [ width (4) | height (4) | bpp (4) ]
 * [ data (width * height * bpp ]
 */

/* set tabstop 8 */

#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <string.h>
#include <errno.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <sys/mman.h>

#include <X11/Xlib.h>
#include <X11/Xutil.h>
#include <X11/Xresource.h>

/*
#define USE_MMAP
 */

static Display *disp;
static Window window;
static GC gc;
static Pixmap pixmap;
static Visual *visual;
static XImage *image;

static int init_visual(int class, int depth)
{
	XVisualInfo tmpl, *vinfo;
	int nvis, mask;

	tmpl.screen = DefaultScreen(disp);
	tmpl.class = class;
	tmpl.depth = depth;
	mask = VisualScreenMask | VisualClassMask | VisualDepthMask;
	vinfo = XGetVisualInfo(disp, mask, &tmpl, &nvis);
	if(!vinfo)
		return -1;
	visual = vinfo[0].visual;
	XFree(vinfo);
	return 0;
}

static int init_window(unsigned int width, unsigned int height, 
	unsigned int depth, int argc, char **argv, char *name)
{
	XSizeHints *xsh;
	XClassHint *xch;
	XWMHints *xwmh;
	XTextProperty wname, iname;
	unsigned long fgpix, bgpix;
	XSetWindowAttributes xswa;

	bgpix = BlackPixel(disp, DefaultScreen(disp));
	fgpix = WhitePixel(disp, DefaultScreen(disp));

	if((xsh = XAllocSizeHints()) == NULL)
		return -1;

	xsh->flags = PSize | PMaxSize | PMinSize;
	xsh->min_width = xsh->max_width = xsh->width = width;
	xsh->min_height = xsh->max_height = xsh->height = height; 

	window = XCreateSimpleWindow(disp, DefaultRootWindow(disp),
		0, 0, /* initial position */ xsh->width, xsh->height, 
		0, /* border width */ fgpix, bgpix);

	pixmap = XCreatePixmap(disp, window, xsh->width, xsh->height, depth);

	if((xch = XAllocClassHint()) == NULL)
		return -1;
	xch->res_name = name;
	xch->res_class = name;

	/* window and icon name */
	if( XStringListToTextProperty(&name, 1, &wname) == 0 ||
		XStringListToTextProperty(&name, 1, &iname) == 0)
		return -1;

	if((xwmh = XAllocWMHints()) == NULL)
		return -1;

	xwmh->flags = (InputHint|StateHint);
	xwmh->input = True;
	xwmh->initial_state = NormalState;
	XSetWMProperties(disp, window, &wname, &iname, argv, argc, xsh, xwmh, 
		xch);
	xswa.colormap = DefaultColormap(disp, DefaultScreen(disp));
	xswa.bit_gravity = CenterGravity;
#if 0
	xswa.event_mask = 	ExposureMask;
#else
	xswa.background_pixmap = pixmap;	/* with this, we don't have to deal
										   with exposure events */
#endif
	XChangeWindowAttributes(disp, window, 
		(CWColormap|CWBitGravity|CWBackPixmap), &xswa);

	/* voila! */
	return 0;	
}

static int init_gc()
{
	unsigned long bgpix, fgpix;
	XGCValues gcv;
	
	bgpix = BlackPixel(disp, DefaultScreen(disp));
	fgpix = WhitePixel(disp, DefaultScreen(disp));
	gcv.foreground = fgpix;
	gcv.background = bgpix;
	gc = XCreateGC(disp, window, (GCForeground | GCBackground), &gcv);

	return 0;
}

static int init_image(unsigned char *data, int width, int height, int depth, 
	int format)
{
	int offset = 0;
	int bitmap_pad = 32;
	int bytes_per_line = 0; /* # of bytes between start of successive scan
				   		       lines. 0 -> lines are contiguous */

	image = XCreateImage(disp, visual, depth, format, offset, data, 
		width, height, bitmap_pad, bytes_per_line);
	if(!image)
			return -1;

	/* depends on the server, maybe should be stored in the header */
	image->byte_order = LSBFirst;

	return 0;
}

static void end_x()
{
	XFreeGC(disp, gc);
	image->data = NULL;
	XDestroyImage(image);
	XFreePixmap(disp, pixmap);
	XDestroyWindow(disp, window);
	XCloseDisplay(disp);
}

static int init_x(int argc, char **argv, unsigned char *data, 
	unsigned int len, unsigned int width, unsigned int height, 
	unsigned int bpp, char *name)
{
	char *display_name = NULL;
	int depth, class, format;

	if( (disp = XOpenDisplay(display_name)) == NULL)
		return -1;

	depth = (bpp == 32 ? 24 : bpp);
	class = TrueColor;
	format = ZPixmap;
	if(init_visual(class, depth) == -1) 
		return -1;
	if(init_window(width, height, depth, argc, argv, name) == -1)
		return -1;
	if(init_gc() == -1)
		return -1;
	if(init_image(data, width, height, depth, format) == -1) 
		return -1;
	XPutImage(disp, pixmap, gc, image, 0, 0, 0, 0, width, height);
	XMapWindow(disp, window);
	XFlush(disp);
	return 0;
}

int main(int argc, char **argv)
{
	int fd, err;
	unsigned int width, height, depth, size, bpp, d;
	unsigned char *data;
	XEvent event;

	if(argc != 2) {
		printf("Usage: %s fn\n", argv[0]);
		return 1;
	}

	fd = open(argv[1], O_RDONLY);
	if(fd == -1) {
		printf("%s: Could not open input file %s: %s\n", argv[0], 
			argv[1], strerror(errno));
		return 1;
	}

	read(fd, &width, 4);
	read(fd, &height, 4);
	read(fd, &bpp, 4);
	read(fd, &d, 4); /* pad */
	width = ntohl(width);
	height = ntohl(height);
	bpp = ntohl(bpp);
	size = width*height*bpp;
	printf("reading image %u*%u*%u. total size: %u\n", width, height, bpp, 
		size);

#ifdef USE_MMAP
	data = mmap(NULL, size, PROT_READ, MAP_SHARED, fd, 0); 
	if(data == (void*)-1) {
		printf("%s: Could not map input file: %s\n", argv[0], 
			strerror(errno));
		close(fd);
		return 1;
	}
#else
	data = malloc(size);
	if(!data) {
		printf("%s: No memory for image data: %s\n", argv[0], 
			strerror(errno));
		close(fd);
		return 1;
	}
	read(fd, data, size);
	close(fd);
#endif

	err = init_x(argc, argv, data, size, width, height, bpp*8, argv[1]);
	if(err == -1) {
		printf("%s: Could not initialize X\n", argv[0]);
#ifdef USE_MMAP
		munmap(data, size);
		close(fd);
#else
		free(data);
#endif
		return 1;
	}

	/* loop until closed */
	while(1) {
		XNextEvent(disp, &event);
		if(event.xany.window != window)
			continue;
	}

	/* the end */
end:
#ifdef USE_MMAP
	munmap(data, size);
	close(fd);
#else
	free(data);
#endif
	end_x();
	printf("c-ya!\n");
	return 0;
}
