/*
 * Application to test X draw operations.
 *
 * Released to the public domain. You may use this
 * code any way you wish, private, educational, or commercial.  It's free.
 *
 * Author:
 * Leo Kim - Network Computing Lab - Columbia University
 */

#include <stdio.h>
#include <stdlib.h>
#include <stdarg.h>		/* for va_args */
#include <time.h>		/* for "struct tm" in msg */
#include <unistd.h>		/* for sleep() */
#include <fcntl.h>		/* for open() */
#include <string.h>		/* for strerror() */
#include <sys/mman.h>		/* for mmap() */
#include <netinet/in.h> 	/* for ntohl() */
#include <errno.h>
#include <X11/Xlib.h>
#include <X11/Xutil.h>

/* for gettimeofday */
#include <sys/time.h>
#include <time.h>

#include <asm/msr.h>

#define gettime_tsc(t) { 	\
	rdtscl((t));		\
}

/* Structures */
struct test_x11 {
	Display *display;
	Window window;
	GC gc;
	Pixmap pixmap;
	Visual *visual;
	XImage *image;
};

/* Defaults */
#define TEST_APP_NAME		"Test Draw"
#define TEST_DEPTH		24
#define TEST_WIDTH		600
#define TEST_HEIGHT		400

/* Image files */
#define TEST_RAW_IMAGE3		"poland.raw"

/* Return values */
#define TEST_OKAY		0
#define TEST_ERROR		-1

#if 0
#define DO_TIMESTAMPS
#endif

#ifdef DO_TIMESTAMPS
#if 0
#define ts(tv)	gettimeofday((tv), NULL)
#else
#define ts(tv) gettime_tsc(tv)
#endif
#else
#define ts(tv)
#endif	/* DO_TIMESTAMPS */

#define DO_SYNC
#ifdef DO_SYNC
#define SYNC()	XSync(x11->display, False)
#else
#define SYNC()
#endif /* DO_SYNC */

/* Initialization prototypes */
static int createWindow(struct test_x11 *x11, int argc, char **argv);
static int getVisual(struct test_x11 *x11);
static int createGC(struct test_x11 *x11);
static int createImage(struct test_x11 *x11, unsigned char *data, int width, 
    int height, int depth, int format);

/* Test functions */
static int testRawImage(struct test_x11 *x11, char *file);
static int testCopy(struct test_x11 *x11);

/* Utility functions */
static void cleanup(struct test_x11 *x11);
static void msg(const char *fmt, ...);

#ifdef DO_TIMESTAMPS
#define NUM_TESTS 30
static unsigned long tv[NUM_TESTS*2];
static unsigned int curr = 0;
#endif

/****************************** Main function ********************************/
int main(int argc, char **argv) 
{
	struct test_x11 *x11;
	Display *disp;
	int retval = TEST_OKAY;
#ifdef DO_TIMESTAMPS
	unsigned int i;
	unsigned long msec, diff;
#endif
	char c;

	if (NULL == (disp = XOpenDisplay(NULL))) {
		msg("Could not initialize display %s. Bailing...\n.", 
			XDisplayName(NULL));
		retval = TEST_ERROR;
		goto cleanup;
	}

	/* init values */
	x11 = malloc(sizeof(struct test_x11));
	x11->display = disp;
	x11->window = (Window)-1;
	x11->gc = (GC)-1;
	x11->pixmap = (Pixmap)-1;
	x11->visual = NULL;
	x11->image = NULL;

	/* initialize visual */
	if (TEST_ERROR == (retval = getVisual(x11))) {
		msg("Could not init Visual.\n");
		goto cleanup;
	}

	/* create window */
	if (TEST_ERROR == (retval = createWindow(x11, argc, argv))) {
		msg("Could not create Window.\n");
		goto cleanup;
	}

	/* create GC */
	if (TEST_ERROR == createGC(x11)) {
		msg("Could not create graphics context.\n");
		goto cleanup;
	}

	/* ok, map the window and we're all set */
	XMapWindow(x11->display, x11->window);
	XFlush(x11->display);

	/* wait for the user to tell us it's ok to start */
	printf("press any key to start the tests...");
	c = getc(stdin);
	printf("i'm off\n");

	/* start the tests */
	if (TEST_ERROR == testRawImage(x11, TEST_RAW_IMAGE3)) {
		msg("An error occured during testRawImage on file %s. "
			"Bailing...\n", TEST_RAW_IMAGE3);
		goto cleanup;
	}

	sleep(1);

	if (TEST_ERROR == testCopy(x11)) {
		msg("An error occured during testCopy. Bailing...\n");
		goto cleanup;
	}

#ifdef DO_TIMESTAMPS
	printf("\n==== results ===\n");
	printf("curr %u\n", curr);
	for(i=0; i<curr; ++i) {
		msec = tv[i];
		++i;
		diff = tv[i];
		printf("test %d\nstart %lu\n", (i/2)+1, msec);
		printf("end %lu\n", diff);
		diff = diff - msec;
		printf("total %lu\n", diff);
	}
#endif

	printf("\ni'm done. press the other any key to finish...");
	c = getc(stdin);
	printf("au revoir!\n");

cleanup:
	cleanup(x11);
	return retval;
}

/**************************** Init functions *********************************/
static int createWindow(struct test_x11 *x11, int argc, char **argv)
{
	const char *func = __func__;
	XSizeHints *xsh;
	XClassHint *xch;
	XWMHints *xwmh; 
	XTextProperty wname, iname;
	XSetWindowAttributes xswa;
	unsigned long fgpix, bgpix;
	char *appname = TEST_APP_NAME;

	bgpix = WhitePixel(x11->display, DefaultScreen(x11->display));
	fgpix = WhitePixel(x11->display, DefaultScreen(x11->display));

	if (NULL == (xsh = XAllocSizeHints())) {
		msg("%s: Error allocating size hints\n", func);
		return TEST_ERROR;
	} 
	xsh->flags = PSize | PMaxSize | PMinSize;
	xsh->min_width = xsh->max_width = xsh->width = TEST_WIDTH;
	xsh->min_height = xsh->max_height = xsh->height = TEST_HEIGHT;

	/* create the window */
	x11->window = XCreateSimpleWindow(x11->display, DefaultRootWindow(x11->display),
		0, 0, /* initial position */
		xsh->width, xsh->height,
		0, /* border width */
		fgpix, bgpix);

	if (NULL == (xch = XAllocClassHint())) {
		msg("%s: Error allocating class hint\n", func);
		return TEST_ERROR;
	}

	xch->res_name = appname;
	xch->res_class = appname;

	/* window and icon name */
	if (XStringListToTextProperty(&appname, 1, &wname) == 0 ||
		XStringListToTextProperty(&appname, 1, &iname) == 0) 
	{
		msg("%s: Error creating XTextProperty\n", func);
		return TEST_ERROR;
	}

	/* XXX: i think this is deprecated in the x world... one day we'll
	 * fix it */
	if (NULL == (xwmh = XAllocWMHints())) {
		msg("%s: Error allocating WM hints\n", func);
		return TEST_ERROR;
	}

	xwmh->flags = (InputHint|StateHint);
	xwmh->input = True;
	xwmh->initial_state = NormalState;
	XSetWMProperties(x11->display, x11->window, &wname, &iname, argv, argc,
		xsh, xwmh, xch);

	xswa.colormap = DefaultColormap(x11->display, DefaultScreen(x11->display));
		xswa.bit_gravity = CenterGravity;

	XChangeWindowAttributes(x11->display, x11->window,
		(CWColormap|CWBitGravity), &xswa);

	return TEST_OKAY;
}

static int getVisual(struct test_x11 *x11)
{
	const char *func = __func__;
	XVisualInfo tmpl, *vinfo;
	int nvis, mask;

	tmpl.screen = DefaultScreen(x11->display);
	tmpl.class = TrueColor;
	tmpl.depth = TEST_DEPTH;
	mask = VisualScreenMask | VisualClassMask | VisualDepthMask;
	vinfo = XGetVisualInfo(x11->display, mask, &tmpl, &nvis);
	if(!vinfo) {
		msg("%s: Couldn't find suitable x11->visual\n", func);
		return TEST_ERROR;
	}
	x11->visual = vinfo[0].visual;
	XFree(vinfo);

	return TEST_OKAY;
}

static int createGC(struct test_x11 *x11)
{
	unsigned long bgpix, fgpix;
	XGCValues gcv;

	bgpix = BlackPixel(x11->display, DefaultScreen(x11->display));
	fgpix = WhitePixel(x11->display, DefaultScreen(x11->display));

	gcv.foreground = fgpix;
	gcv.background = bgpix;

	x11->gc = XCreateGC(x11->display, x11->window,
	(GCForeground | GCBackground), &gcv);

	return TEST_OKAY;
}

static int createImage(struct test_x11 *x11, unsigned char *data, int width, 
    int height, int depth, int format)
{
	int offset = 0;
	int bitmap_pad = 32;
	int bytes_per_line = 0; /* # of bytes between start of successive scan
							   lines. 0 -> lines are contiguous */

	x11->image = XCreateImage(x11->display, x11->visual, depth, format, offset, data,
		width, height, bitmap_pad, bytes_per_line);

	if(!x11->image) return TEST_ERROR;

	x11->image->byte_order = LSBFirst;

	printf("%dx%d\n", width, height);
	x11->pixmap = XCreatePixmap(x11->display, x11->window, width,
		height, TEST_DEPTH);
	sleep(1);

	return TEST_OKAY;
}

/************************** Test functions *********************************/
static int testRawImage(struct test_x11 *x11, char *file)
{
	int fd;
	char *data;
	int width, height, bpp, size, depth;
	int retval = TEST_OKAY;

	if(-1 == (fd = open(file, O_RDONLY))) {
		msg("Could not open input file %s. %s\n", strerror(errno));
		retval = TEST_ERROR;
		goto end_raw;
	} 

	read(fd, &width, 4);
	read(fd, &height, 4);
	read(fd, &bpp, 4);
	width = ntohl(width);
	height = ntohl(height);
	bpp = ntohl(bpp);
	size = width*height*bpp;

	bpp *= 8;
	depth = (bpp == 32 ? 24 : bpp);

	data = mmap(NULL, size, PROT_READ, MAP_SHARED, fd, 0);
	if((void*)-1 == data) {
		msg("Could not map input file: %s\n", strerror(errno));
		close(fd);
		retval = TEST_ERROR;
		goto end_raw;
	}

	if(TEST_ERROR == (retval = createImage(x11, data, width, height, depth, 
			ZPixmap))) {
		goto cleanup_raw;
	}
	SYNC();
	ts(tv[curr++]);

	XPutImage(x11->display, x11->pixmap, x11->gc, x11->image, 0, 0, 0, 0, 
		width, height);
	XCopyArea(x11->display, x11->pixmap, x11->window, x11->gc, 0, 0, 
		TEST_WIDTH, TEST_HEIGHT, 0, 0);

#if 0
	XFlush(x11->display);
#endif
	SYNC();
	ts(tv[curr++]);

cleanup_raw:
	munmap(data, size);
	close(fd);

end_raw:
	return retval;
}

static int testCopy(struct test_x11 *x11)
{
	int i;
	char c;
	const int tot = x11->image->height-TEST_HEIGHT;
again:
	ts(tv[curr++]);
	for(i=0; i<tot; i+=2) {
		XCopyArea(x11->display, x11->window, x11->window, x11->gc, 
			0, 2, 
			TEST_WIDTH, TEST_HEIGHT-2, 
			0, 0);
		XCopyArea(x11->display, x11->pixmap, x11->window, x11->gc, 
			0, TEST_HEIGHT+i, 
			TEST_WIDTH, 2, 
			0, TEST_HEIGHT-2);
	}
	SYNC();
	ts(tv[curr++]);
	ts(tv[curr++]);
	for(i-=2; i>=0; i-=2) {
		XCopyArea(x11->display, x11->window, x11->window, x11->gc, 
			0, 0,
			TEST_WIDTH, TEST_HEIGHT-2, 
			0, 2);
		XCopyArea(x11->display, x11->pixmap, x11->window, x11->gc, 
			0, i, 
			TEST_WIDTH, 2, 
			0, 0);
	}
	SYNC();
	ts(tv[curr++]);

	printf("repeat? [Y/n] ");
	c = getc(stdin);
	if(c != 'n')
		goto again;

	return TEST_OKAY;
}

/************************* Utility functions ********************************/
static void cleanup(struct test_x11 *x11) 
{
	if((GC)-1 != x11->gc) XFreeGC(x11->display, x11->gc);
	if(x11->image) {
		/* Strange: setting data to NULL prevents XDestroyImage from 
		 * segfaulting. */
		x11->image->data = NULL;
		XDestroyImage(x11->image);
	}
	if((Window)-1 != x11->window) XDestroyWindow(x11->display, x11->window);
	if((Pixmap)-1 != x11->pixmap) XFreePixmap(x11->display, x11->pixmap);

	XCloseDisplay(x11->display);

	free(x11);
}

static void msg(const char *fmt, ...)
{
	va_list v;
	time_t t;
	struct tm b;

	time(&t);	    
	localtime_r(&t, &b);

	va_start(v, fmt);
	vfprintf(stdout, fmt, v);
	va_end(v);
}

