/* Copyright 2005, 2006, 2007 Ricardo Baratto */
/* Copyright 2005, 2006, 2007 Columbia University */
/* This file is part of THINC.

    THINC is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 2 of the License, or
    (at your option) any later version.

    THINC is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with THINC.  If not, see <http://www.gnu.org/licenses/>.
*/
/*
 * Program to play with (test?) THINC's X Extension
 *
 * Authors: 
 * Ricardo A. Baratto - Network Computing Lab - Columbia University
 */

/* set tabstop 8 */

#include <stdio.h>
#include <stdlib.h>
#include <X11/X.h>
#include <X11/Xlib.h>
#include <X11/Xutil.h>
#include <X11/Xresource.h>
#include <X11/extensions/xthinc.h>
#include <X11/extensions/xthincstr.h>
#include <getopt.h>
#include <string.h>

#define T_SET_DEBUG_LEVEL	0x00000001
#define	T_STATS_FLUSH		0x00000002
#define	T_STATS_RESET		0x00000004
#define	T_INSTR_FLUSH		0x00000010
#define	T_INSTR_RESET		0x00000020
#define T_WINDOW_SHARE		0x00000040

static int be_verbose = 0;
#define verbose(...)	if(be_verbose) printf(__VA_ARGS__)

static void usage(const char *name);
static void print_debug_levels();
static int parse_debug_level(const char *arg);
static int set_debug_level(Display *disp, int debug_level);
static int stats_flush(Display *disp);
static int stats_reset(Display *disp);
static int instr_flush(Display *disp);
static int instr_reset(Display *disp);
static int window_share(Display *disp, const char* win);

int main(int argc, char **argv)
{
	Display *disp;
	int ver_maj, ver_min;
	int opt, index = 0;
	struct option long_options[] = {
		{"help", 0, 0, 'h'},
		{"verbose", 0, 0, 'v'},
		{"debug", 1, 0, 'd'},
		{"print", 0, 0, 'p'},
		{"stats-flush", 0, 0, 'f'},
		{"stats-reset", 0, 0, 'r'},
		{"instr-flush", 0, 0, 'i'},
		{"instr-reset", 0, 0, 's'},
		{"share-window", 1, 0, 'w'},
		{0,0,0,0}
	};
	int debug_level = -1;
	char *window = NULL;
	unsigned int oper = 0;

	while(1) {
		opt = getopt_long(argc, argv, "hvd:pfrisw:", long_options, &index);
		if(opt == -1)
			/* done */
			break;
		switch(opt) {
			case 'h':
				usage(argv[0]);
				return EXIT_SUCCESS;
			case 'v':
				be_verbose = 1;
				break;
			case 'd':
				debug_level = parse_debug_level(optarg);
				oper |= T_SET_DEBUG_LEVEL;
				break;
			case 'p':
				print_debug_levels();
				return EXIT_SUCCESS;
			case 'f':
				oper |= T_STATS_FLUSH;
				break;
			case 'r':
				oper |= T_STATS_RESET;
				break;
			case 'i':
				oper |= T_INSTR_FLUSH;
				break;
			case 's':
				oper |= T_INSTR_RESET;
				break;
			case 'w':
				oper |= T_WINDOW_SHARE;
				window = optarg;
				break;
			default:
				usage(argv[0]);
				return EXIT_FAILURE;
		}
	}

	if(!oper) {
		usage(argv[0]);
		return EXIT_FAILURE;
	}

	if(!(disp = XOpenDisplay(NULL)) ) {
		printf("no display man. what's up with that?\n");
		return EXIT_FAILURE;
	}

	verbose("is THINC in %s? ", XDisplayName(NULL));

	if(!XTHINCQueryExtension(disp)) {
		printf("dsol monsieur, kein THINC at display %s\n",
			XDisplayName(NULL));
		XCloseDisplay(disp);
		return EXIT_FAILURE;
	}

	verbose("oui monsieur!\n");
	verbose("which version? ");
	if(!XTHINCQueryVersion(disp, &ver_maj, &ver_min)) {
		printf("crapers, no version :(\n");
		XCloseDisplay(disp);
		return EXIT_FAILURE;
	}
	verbose("why, none other than %d.%d!\n", ver_maj, ver_min);

	/* now see what the user wants us to do */
	if(oper & T_SET_DEBUG_LEVEL) {
		if(set_debug_level(disp, debug_level) == EXIT_FAILURE) {
			XCloseDisplay(disp);
			return EXIT_FAILURE;
		}
	}
	if(oper & T_STATS_FLUSH) {
		if(stats_flush(disp) == EXIT_FAILURE) {
			XCloseDisplay(disp);
			return EXIT_FAILURE;
		}
	}
	if(oper & T_STATS_RESET) {
		if(stats_reset(disp) == EXIT_FAILURE) {
			XCloseDisplay(disp);
			return EXIT_FAILURE;
		}
	}
	if(oper & T_INSTR_FLUSH) {
		if(instr_flush(disp) == EXIT_FAILURE) {
			XCloseDisplay(disp);
			return EXIT_FAILURE;
		}
	}
	if(oper & T_INSTR_RESET) {
		if(instr_reset(disp) == EXIT_FAILURE) {
			XCloseDisplay(disp);
			return EXIT_FAILURE;
		}
	}
	if(oper & T_WINDOW_SHARE) {
		if(window_share(disp, window) == EXIT_FAILURE) {
			XCloseDisplay(disp);
			return EXIT_FAILURE;
		}
	}
	XSync(disp, False);

	verbose("au revoir!\n");
	XCloseDisplay(disp);
	return EXIT_SUCCESS;
}

static void usage(const char *name)
{
	printf("usage: %s <operation> [options]\n", name);
	printf("operations:\n"
		"--debug/-d            set server's debug level\n"
		"--stats-flush/-f      tell server to write its stats to the\n"
		"                      log file\n"
		"--stats-reset/-r      tell server to reset its stats\n"
		"--instr-flush/-i      tell server to write its\n"
		"                      instrumentation to the log file\n"
		"--instr-reset/-s      tell server to reset its\n"
		"                      instrumentation\n"
		"--share-window/-w     tell server to share a window. you can\n"
		"                      get the window id using xwininfo\n"
		"\n"
	);
	printf("options:\n"
		"--help/-h             show this message\n"
		"--verbose/-v          be verbose\n"
		"--print/-p            print debug levels\n"
		"\n"
	);
}

static void print_debug_levels()
{
	printf("server debug levels:\n"
		"- %-20s %#010x\n"
		"- %-20s %#010x\n"
		"- %-20s %#010x\n"
		"- %-20s %#010x\n"
		"- %-20s %#010x\n"
		"- %-20s %#010x\n"
		"- %-20s %#010x\n"
		"- %-20s %#010x\n"
		"- %-20s %#010x\n"
		"- %-20s %#010x\n"
		"- %-20s %#010x\n"
		"- %-20s %#010x\n"
		"- %-20s %#010x\n"
		"- %-20s %#010x\n"
		"- %-20s %#010x\n"
		"- %-20s %#010x\n",
		"T_DBG_ERROR", 0x0001,
		"T_DBG_BASE", 0x0002,
		"T_DBG_CLIENT", 0x0004,
		"T_DBG_IO", 0x0008,
		"T_DBG_PROTO", 0x0010,
		"T_DBG_STATS", 0x0020,
		"T_DBG_OTHER", 0x0040,
		"UNUSED", 0x0080,
		"T_DBG_DRAW", 0x0100,
		"T_DBG_WIN", 0x0200,
		"T_DBG_CMD", 0x0400,
		"T_DBG_INPUT", 0x0800,
		"T_DBG_VID", 0x1000,
		"T_DBG_COMP", 0x2000,
		"T_DBG_TMP", 0x4000,
		"T_DBG_ANNOY", 0x8000
	);
}

static int parse_debug_level(const char *arg)
{
	unsigned int debug_level;
	if(*arg == '0' && *(arg+1) == 'x')
		sscanf(arg, "%x", &debug_level);
	else
		sscanf(arg, "%d", &debug_level);

	return debug_level;
}

static int set_debug_level(Display *disp, int debug_level)
{
	int debug;
	verbose("setting debug level to %#x ... ", debug_level);
	XTHINCSetValue(disp, xTHINC_val_debugLevel, debug_level);
	/* now get it to make sure 'tis good */
	XTHINCGetValue(disp, xTHINC_val_debugLevel, &debug);
	if(debug != debug_level) {
		printf("scheisse, got different debug level :( "
			"%d vs %d\n", debug_level, debug);
		return EXIT_FAILURE;
	}
	printf("wazam!\n");
	return EXIT_SUCCESS;
}

static int stats_flush(Display *disp)
{
	verbose("telling server to flush its stats...\n");
	XTHINCStatsFlush(disp);
	return EXIT_SUCCESS;
}

static int stats_reset(Display *disp)
{
	verbose("telling server to reset its stats...\n");
	XTHINCStatsReset(disp);
	return EXIT_SUCCESS;
}

static int instr_flush(Display *disp)
{
	verbose("telling server to flush its instrumentation...\n");
	XTHINCInstrFlush(disp);
	return EXIT_SUCCESS;
}

static int instr_reset(Display *disp)
{
	verbose("telling server to reset its instrumentation...\n");
	XTHINCInstrReset(disp);
	return EXIT_SUCCESS;
}

static int window_share(Display *disp, const char *win)
{
	Window w;
	verbose("sharing window %s...", win);

	/* from appshare-x11 by jonathan lennox */
	if (strcmp(win, "desktop") == 0) {
		w = XRootWindow(disp, XDefaultScreen(disp));
	} else {
		w = 0;
		sscanf(win, "0x%lx", &w);
		if (w == 0) {
			sscanf(win, "%lu", &w);
		}
		if (w == 0) {
			printf("invalid window %s\n", win);
			return EXIT_FAILURE;
		}
	}
	if(!XTHINCWindowShare(disp, w)) {
		printf("Couldn't share window %s\n", win);
		return EXIT_FAILURE;
	}
	verbose("finished\n");
	return EXIT_SUCCESS;
}
